import os
import struct

import numpy as np
import pytest
from numpy.testing import assert_equal

from nengo.exceptions import CacheIOError
from nengo.utils import nco
from nengo.utils.nco import Subfile


@pytest.fixture(name="data")
def fixture_data():
    return "0123456789\n123456789"


@pytest.fixture(name="testfile")
def fixture_testfile(data, tmp_path):
    f = tmp_path / "file.txt"
    f.write_text(data)
    return f


def write_custom_nco_header(fileobj, magic_string="NCO", version=0):
    magic_string = magic_string.encode("utf-8")
    header_format = f"@{len(magic_string)}sBLLLL"
    assert struct.calcsize(header_format) == nco.HEADER_SIZE

    header = struct.pack(header_format, magic_string, version, 0, 1, 2, 3)
    fileobj.write(header)


class TestSubfile:
    def test_reads_only_from_start_to_end(self, data, testfile):
        with testfile.open() as f:
            assert Subfile(f, 2, 6).read() == data[2:6]
        with testfile.open() as f:
            assert Subfile(f, 2, 6).read(8) == data[2:6]
        with testfile.open() as f:
            assert Subfile(f, 2, 6).readline() == data[2:6]
        with testfile.open() as f:
            assert Subfile(f, 2, 6).readline(8) == data[2:6]

    def test_respects_read_size(self, data, testfile):
        with testfile.open() as f:
            assert Subfile(f, 2, 6).read(2) == data[2:4]
        with testfile.open() as f:
            assert Subfile(f, 2, 6).readline(2) == data[2:4]

    def test_readline(self, data, testfile):
        with testfile.open() as f:
            assert Subfile(f, 2, 14).readline() == data[2:11]
        with testfile.open() as f:
            assert Subfile(f, 2, 14).readline(15) == data[2:11]

    def test_readinto(self, data, testfile):
        b = bytearray(4)
        with testfile.open("rb") as f:
            assert Subfile(f, 2, 6).readinto(b) == 4
        assert b == b"2345"

    def test_seek(self, data, testfile):
        with testfile.open() as f:
            sf = Subfile(f, 2, 6)
            sf.seek(2)
            assert sf.read() == data[4:6]
        with testfile.open() as f:
            sf = Subfile(f, 2, 6)
            sf.read(1)
            sf.seek(2, os.SEEK_CUR)
            assert sf.read() == data[5:6]
        with testfile.open() as f:
            sf = Subfile(f, 2, 6)
            sf.seek(-2, os.SEEK_END)
            assert sf.read() == data[4:6]

    def test_seek_before_start(self, data, testfile):
        with testfile.open() as f:
            sf = Subfile(f, 2, 6)
            sf.seek(-2)
            assert sf.read() == data[2:6]
        with testfile.open() as f:
            sf = Subfile(f, 2, 6)
            sf.read(1)
            sf.seek(-4, os.SEEK_CUR)
            assert sf.read() == data[2:6]
        with testfile.open() as f:
            sf = Subfile(f, 2, 6)
            sf.seek(-8, os.SEEK_END)
            assert sf.read() == data[2:6]

    def test_seek_after_end(self, testfile):
        with testfile.open() as f:
            sf = Subfile(f, 2, 6)
            sf.seek(8)
            assert sf.read() == ""
        with testfile.open() as f:
            sf = Subfile(f, 2, 6)
            sf.read(1)
            sf.seek(8, os.SEEK_CUR)
            assert sf.read() == ""
        with testfile.open() as f:
            sf = Subfile(f, 2, 6)
            sf.seek(8, os.SEEK_END)
            assert sf.read() == ""

    def test_tell(self, data, testfile):
        with testfile.open() as f:
            sf = Subfile(f, 2, 6)
            assert sf.tell() == 0
            sf.seek(3)
            assert sf.tell() == 3

    def test_read_errors(self, tmp_path):
        # use a bad magic string
        filepath = tmp_path / "bad_magic_cache_file.txt"
        with open(filepath, "wb") as fh:
            write_custom_nco_header(fh, magic_string="BAD")

        with open(filepath, "rb") as fh:
            with pytest.raises(CacheIOError, match="Not a Nengo cache object file"):
                nco.read(fh)

        # use a bad version number
        filepath = tmp_path / "bad_version_cache_file.txt"
        with open(filepath, "wb") as fh:
            write_custom_nco_header(fh, version=255)

        with open(filepath, "rb") as fh:
            with pytest.raises(CacheIOError, match="NCO protocol version 255 is"):
                nco.read(fh)


def test_nco_roundtrip(tmp_path):
    tmpfile = tmp_path / "test.nco"

    pickle_data = {"0": 237, "str": "foobar"}
    array = np.array([[4, 3], [2, 1]])

    with tmpfile.open("wb") as f:
        nco.write(f, pickle_data, array)

    with tmpfile.open("rb") as f:
        pickle_data2, array2 = nco.read(f)

    assert pickle_data == pickle_data2
    assert_equal(array, array2)



from nengo.network import Network


class Module(Network):
    """Base class for SPA Modules.

    Modules are networks that also have a list of inputs and outputs,
    each with an associated `.Vocabulary` (or a desired dimensionality for
    the vocabulary).

    The inputs and outputs are dictionaries that map a name to an
    (object, Vocabulary) pair. The object can be a `.Node` or an `.Ensemble`.
    """

    def __init__(self, label=None, seed=None, add_to_container=None):
        super().__init__(label, seed, add_to_container)
        self.inputs = {}
        self.outputs = {}

    def on_add(self, spa):
        """Called when this is assigned to a variable in the SPA network.

        Overload this when you want processing to be delayed until after
        the module is attached to the SPA network. This is usually for
        modules that connect to other things in the SPA model (such as
        the basal ganglia or thalamus).
        """



import warnings

import numpy as np

from nengo.config import Config
from nengo.connection import Connection
from nengo.dists import Choice, Uniform
from nengo.ensemble import Ensemble
from nengo.exceptions import ObsoleteError
from nengo.network import Network
from nengo.networks.ensemblearray import EnsembleArray
from nengo.node import Node
from nengo.solvers import NnlsL2nz
from nengo.synapses import Lowpass


# connection weights from (Gurney, Prescott, & Redgrave, 2001)
class Weights:
    mm = 1
    mp = 1
    me = 1
    mg = 1
    ws = 1
    wt = 1
    wm = 1
    wg = 1
    wp = 0.9
    we = 0.3
    e = 0.2
    ep = -0.25
    ee = -0.2
    eg = -0.2
    le = 0.2
    lg = 0.2

    @classmethod
    def str_func(cls, x):
        if x < cls.e:
            return 0
        return cls.mm * (x - cls.e)

    @classmethod
    def stn_func(cls, x):
        if x < cls.ep:
            return 0
        return cls.mp * (x - cls.ep)

    @classmethod
    def gpe_func(cls, x):
        if x < cls.ee:
            return 0
        return cls.me * (x - cls.ee)

    @classmethod
    def gpi_func(cls, x):
        if x < cls.eg:
            return 0
        return cls.mg * (x - cls.eg)


def config_with_default_synapse(config, synapse):
    if config is None:
        config = Config(Connection)
        config[Connection].synapse = synapse
    override = "synapse" not in config[Connection]
    if override:
        config[Connection].synapse = synapse
    return config, override


class BasalGanglia(Network):
    """Winner take all network, typically used for action selection.

    The basal ganglia network outputs approximately 0 at the dimension with
    the largest value, and is negative elsewhere.

    While the basal ganglia is primarily defined by its winner-take-all
    function, it is also organized to match the organization of the human
    basal ganglia. It consists of five ensembles:

    * Striatal D1 dopamine-receptor neurons (``strD1``)
    * Striatal D2 dopamine-receptor neurons (``strD2``)
    * Subthalamic nucleus (``stn``)
    * Globus pallidus internus / substantia nigra reticulata (``gpi``)
    * Globus pallidus externus (``gpe``)

    Interconnections between these areas are also based on known
    neuroanatomical connections. See [1]_ for more details, and [2]_ for
    the original non-spiking basal ganglia model by
    Gurney, Prescott & Redgrave that this model is based on.

    .. note:: The default `.Solver` for the basal ganglia is `.NnlsL2nz`, which
              requires SciPy. If SciPy is not installed, the global default
              solver will be used instead.

    Parameters
    ----------
    dimensions : int
        Number of dimensions (i.e., actions).
    n_neurons_per_ensemble : int, optional
        Number of neurons in each ensemble in the network.
    output_weight : float, optional
        A scaling factor on the output of the basal ganglia
        (specifically on the connection out of the GPi).
    input_bias : float, optional
        An amount by which to bias all dimensions of the input node.
        Biasing the input node is important for ensuring that all input
        dimensions are positive and easily comparable.
    ampa_config : config, optional
        Configuration for connections corresponding to biological connections
        to AMPA receptors (i.e., connections from STN to to GPi and GPe).
        If None, a default configuration using a 2 ms lowpass synapse
        will be used.
    gaba_config : config, optional
        Configuration for connections corresponding to biological connections
        to GABA receptors (i.e., connections from StrD1 to GPi, StrD2 to GPe,
        and GPe to GPi and STN). If None, a default configuration using an
        8 ms lowpass synapse will be used.
    **kwargs
        Keyword arguments passed through to ``nengo.Network``
        like 'label' and 'seed'.

    Attributes
    ----------
    bias_input : Node or None
        If ``input_bias`` is non-zero, this node will be created to bias
        all of the dimensions of the input signal.
    gpe : EnsembleArray
        Globus pallidus externus ensembles.
    gpi : EnsembleArray
        Globus pallidus internus ensembles.
    input : Node
        Accepts the input signal.
    output : Node
        Provides the output signal.
    stn : EnsembleArray
        Subthalamic nucleus ensembles.
    strD1 : EnsembleArray
        Striatal D1 ensembles.
    strD2 : EnsembleArray
        Striatal D2 ensembles.

    References
    ----------
    .. [1] Stewart, T. C., Choo, X., & Eliasmith, C. (2010).
       Dynamic behaviour of a spiking model of action selection in the
       basal ganglia. In Proceedings of the 10th international conference on
       cognitive modeling (pp. 235-40).
    .. [2] Gurney, K., Prescott, T., & Redgrave, P. (2001).
       A computational model of action selection in the basal
       ganglia. Biological Cybernetics 84, 401-423.
    """

    def __init__(
        self,
        dimensions,
        n_neurons_per_ensemble=100,
        output_weight=-3.0,
        input_bias=0.0,
        ampa_config=None,
        gaba_config=None,
        **kwargs,
    ):
        if "net" in kwargs:
            raise ObsoleteError("The 'net' argument is no longer supported.")
        kwargs.setdefault("label", "Basal Ganglia")
        super().__init__(**kwargs)

        ampa_config, override_ampa = config_with_default_synapse(
            ampa_config, Lowpass(0.002)
        )
        gaba_config, override_gaba = config_with_default_synapse(
            gaba_config, Lowpass(0.008)
        )

        # Affects all ensembles / connections in the BG
        # unless they've been overridden on `self.config`
        config = Config(Ensemble, Connection)
        config[Ensemble].radius = 1.5
        config[Ensemble].encoders = Choice([[1]])
        try:
            # Best, if we have SciPy
            config[Connection].solver = NnlsL2nz()
        except ImportError:
            # Warn if we can't use the better decoder solver.
            warnings.warn(
                "SciPy is not installed, so BasalGanglia will "
                "use the default decoder solver. Installing SciPy "
                "may improve BasalGanglia performance."
            )

        ea_params = {"n_neurons": n_neurons_per_ensemble, "n_ensembles": dimensions}

        with self, config:
            self.strD1 = EnsembleArray(
                label="Striatal D1 neurons",
                intercepts=Uniform(Weights.e, 1),
                **ea_params,
            )
            self.strD2 = EnsembleArray(
                label="Striatal D2 neurons",
                intercepts=Uniform(Weights.e, 1),
                **ea_params,
            )
            self.stn = EnsembleArray(
                label="Subthalamic nucleus",
                intercepts=Uniform(Weights.ep, 1),
                **ea_params,
            )
            self.gpi = EnsembleArray(
                label="Globus pallidus internus",
                intercepts=Uniform(Weights.eg, 1),
                **ea_params,
            )
            self.gpe = EnsembleArray(
                label="Globus pallidus externus",
                intercepts=Uniform(Weights.ee, 1),
                **ea_params,
            )

            self.input = Node(label="input", size_in=dimensions)
            self.output = Node(label="output", size_in=dimensions)

            # add bias input (BG performs best in the range 0.5--1.5)
            if abs(input_bias) > 0.0:
                self.bias_input = Node(
                    np.ones(dimensions) * input_bias, label="basal ganglia bias"
                )
                Connection(self.bias_input, self.input)

            # spread the input to StrD1, StrD2, and STN
            Connection(
                self.input,
                self.strD1.input,
                synapse=None,
                transform=Weights.ws * (1 + Weights.lg),
            )
            Connection(
                self.input,
                self.strD2.input,
                synapse=None,
                transform=Weights.ws * (1 - Weights.le),
            )
            Connection(self.input, self.stn.input, synapse=None, transform=Weights.wt)

            # connect the striatum to the GPi and GPe (inhibitory)
            strD1_output = self.strD1.add_output("func_str", Weights.str_func)
            strD2_output = self.strD2.add_output("func_str", Weights.str_func)
            with gaba_config:
                Connection(strD1_output, self.gpi.input, transform=-Weights.wm)
                Connection(strD2_output, self.gpe.input, transform=-Weights.wm)

            # connect the STN to GPi and GPe (broad and excitatory)
            tr = Weights.wp * np.ones((dimensions, dimensions))
            stn_output = self.stn.add_output("func_stn", Weights.stn_func)
            with ampa_config:
                Connection(stn_output, self.gpi.input, transform=tr)
                Connection(stn_output, self.gpe.input, transform=tr)

            # connect the GPe to GPi and STN (inhibitory)
            gpe_output = self.gpe.add_output("func_gpe", Weights.gpe_func)
            with gaba_config:
                Connection(gpe_output, self.gpi.input, transform=-Weights.we)
                Connection(gpe_output, self.stn.input, transform=-Weights.wg)

            # connect GPi to output (inhibitory)
            gpi_output = self.gpi.add_output("func_gpi", Weights.gpi_func)
            Connection(gpi_output, self.output, synapse=None, transform=output_weight)

        # Return ampa_config and gaba_config to previous states, if changed
        if override_ampa:
            del ampa_config[Connection].synapse
        if override_gaba:
            del gaba_config[Connection].synapse


class Thalamus(Network):
    """Inhibits non-selected actions.

    The thalamus is intended to work in tandem with a basal ganglia network.
    It converts basal ganglia output into a signal with (approximately) 1 for
    the selected action and 0 elsewhere.

    In order to suppress low responses and strengthen high responses,
    a constant bias is added to each dimension (i.e., action), and dimensions
    mutually inhibit each other. Additionally, the ensemble representing
    each dimension is created with positive encoders and can be assigned
    positive x-intercepts to threshold low responses.

    Parameters
    ----------
    dimensions : int
        Number of dimensions (i.e., actions).
    n_neurons_per_ensemble : int, optional
        Number of neurons in each ensemble in the network.
    mutual_inhib : float, optional
        Strength of the mutual inhibition between actions.
    threshold : float, optional
        The threshold below which values will not be represented.
    **kwargs
        Keyword arguments passed through to ``nengo.Network``
        like 'label' and 'seed'.

    Attributes
    ----------
    actions : EnsembleArray
        Each ensemble represents one dimension (action).
    bias : Node
        The constant bias injected in each ``actions`` ensemble.
    input : Node
        Input to the ``actions`` ensembles.
    output : Node
        Output from the ``actions`` ensembles.
    """

    def __init__(
        self,
        dimensions,
        n_neurons_per_ensemble=50,
        mutual_inhib=1.0,
        threshold=0.0,
        **kwargs,
    ):
        if "net" in kwargs:
            raise ObsoleteError("The 'net' argument is no longer supported.")
        kwargs.setdefault("label", "Thalamus")
        super().__init__(**kwargs)

        with self:
            self.actions = EnsembleArray(
                n_neurons_per_ensemble,
                dimensions,
                intercepts=Uniform(threshold, 1),
                encoders=Choice([[1.0]]),
                label="actions",
            )
            Connection(
                self.actions.output,
                self.actions.input,
                transform=(np.eye(dimensions) - 1) * mutual_inhib,
            )
            self.bias = Node([1], label="thalamus bias")
            Connection(
                self.bias, self.actions.input, transform=np.ones((dimensions, 1))
            )

        self.input = self.actions.input
        self.output = self.actions.output



import warnings
from copy import deepcopy

from nengo.config import Config
from nengo.connection import Connection
from nengo.ensemble import Ensemble
from nengo.exceptions import (
    ConfigError,
    NetworkContextError,
    NotAddedToNetworkWarning,
    ReadonlyError,
)
from nengo.node import Node
from nengo.params import IntParam, StringParam
from nengo.probe import Probe
from nengo.utils.threading import ThreadLocalStack


class Network:
    """A network contains ensembles, nodes, connections, and other networks.

    A network is primarily used for grouping together related
    objects and connections for visualization purposes.
    However, you can also use networks as a nice way to reuse
    network creation code.

    To group together related objects that you do not need to reuse,
    you can create a new ``Network`` and add objects in a ``with`` block.
    For example:

    .. testcode::

       network = nengo.Network()
       with network:
           with nengo.Network(label="Vision"):
               v1 = nengo.Ensemble(n_neurons=100, dimensions=2)
           with nengo.Network(label="Motor"):
               sma = nengo.Ensemble(n_neurons=100, dimensions=2)
           nengo.Connection(v1, sma)

    To reuse a group of related objects, you can create a new subclass
    of ``Network``, and add objects in the ``__init__`` method.
    For example:

    .. testcode::

       class OcularDominance(nengo.Network):
           def __init__(self):
               self.column = nengo.Ensemble(n_neurons=100, dimensions=2)

       network = nengo.Network()
       with network:
           left_eye = OcularDominance()
           right_eye = OcularDominance()
           nengo.Connection(left_eye.column, right_eye.column)

    Parameters
    ----------
    label : str, optional
        Name of the network.
    seed : int, optional
        Random number seed that will be fed to the random number generator.
        Setting the seed makes the network's build process deterministic.
    add_to_container : bool, optional
        Determines if this network will be added to the current container.
        If None, this network will be added to the network at the top of the
        ``Network.context`` stack unless the stack is empty.

    Attributes
    ----------
    connections : list
        `.Connection` instances in this network.
    ensembles : list
        `.Ensemble` instances in this network.
    label : str
        Name of this network.
    networks : list
        `.Network` instances in this network.
    nodes : list
        `.Node` instances in this network.
    probes : list
        `.Probe` instances in this network.
    seed : int
        Random seed used by this network.
    """

    context = ThreadLocalStack(maxsize=100)  # static stack of Network objects

    label = StringParam("label", optional=True, readonly=False)
    seed = IntParam("seed", optional=True, readonly=False)

    def __init__(self, label=None, seed=None, add_to_container=None):
        self.label = label
        self.seed = seed
        self._config = self.default_config()

        self._objects = {Ensemble: [], Node: [], Connection: [], Network: [], Probe: []}
        self._ensembles = self.objects[Ensemble]
        self._nodes = self.objects[Node]
        self._connections = self.objects[Connection]
        self._networks = self.objects[Network]
        self._probes = self.objects[Probe]

        # By default, we want to add to the current context, unless there is
        # no context; i.e., we're creating a top-level network.
        if add_to_container is None:
            add_to_container = len(Network.context) > 0

        if add_to_container:
            Network.add(self)

    @staticmethod
    def add(obj):
        """Add the passed object to ``Network.context``."""
        if len(Network.context) == 0:
            raise NetworkContextError(
                f"'{obj}' must either be created inside a ``with network:`` "
                "block, or set add_to_container=False in the object's "
                "constructor."
            )
        network = Network.context[-1]
        if not isinstance(network, Network):
            raise NetworkContextError(f"Current context ({network}) is not a network")
        for cls in type(obj).__mro__:
            if cls in network.objects:
                network.objects[cls].append(obj)
                break
        else:
            raise NetworkContextError(
                f"Objects of type '{type(obj).__name__}' cannot be added to networks."
            )

    @staticmethod
    def default_config():
        """Constructs a `~.Config` object for setting defaults."""
        return Config(Connection, Ensemble, Node, Probe)

    def _all_objects(self, object_type):
        """Returns a list of all objects of the specified type."""
        # Make a copy of this network's list
        objects = list(self.objects[object_type])
        for subnet in self.networks:
            objects.extend(subnet._all_objects(object_type))
        return objects

    @property
    def all_objects(self):
        """(list) All objects in this network and its subnetworks."""
        objects = []
        for object_type in self.objects:
            objects.extend(self._all_objects(object_type))
        return objects

    @property
    def all_ensembles(self):
        """(list) All ensembles in this network and its subnetworks."""
        return self._all_objects(Ensemble)

    @property
    def all_nodes(self):
        """(list) All nodes in this network and its subnetworks."""
        return self._all_objects(Node)

    @property
    def all_networks(self):
        """(list) All networks in this network and its subnetworks."""
        return self._all_objects(Network)

    @property
    def all_connections(self):
        """(list) All connections in this network and its subnetworks."""
        return self._all_objects(Connection)

    @property
    def all_probes(self):
        """(list) All probes in this network and its subnetworks."""
        return self._all_objects(Probe)

    @property
    def objects(self):
        return self._objects

    @objects.setter
    def objects(self, _):
        raise ReadonlyError(attr="objects", obj=self)

    @property
    def ensembles(self):
        return self._ensembles

    @ensembles.setter
    def ensembles(self, _):
        raise ReadonlyError(attr="ensembles", obj=self)

    @property
    def nodes(self):
        return self._nodes

    @nodes.setter
    def nodes(self, _):
        raise ReadonlyError(attr="nodes", obj=self)

    @property
    def networks(self):
        return self._networks

    @networks.setter
    def networks(self, _):
        raise ReadonlyError(attr="networks", obj=self)

    @property
    def connections(self):
        return self._connections

    @connections.setter
    def connections(self, _):
        raise ReadonlyError(attr="connections", obj=self)

    @property
    def probes(self):
        return self._probes

    @probes.setter
    def probes(self, _):
        raise ReadonlyError(attr="probes", obj=self)

    @property
    def config(self):
        """(`.Config`) Configuration for this network."""
        return self._config

    @config.setter
    def config(self, _):
        raise ReadonlyError(attr="config", obj=self)

    @property
    def n_neurons(self):
        """(int) Number of neurons in this network, including subnetworks."""
        return sum(ens.n_neurons for ens in self.all_ensembles)

    def __contains__(self, obj):
        return type(obj) in self.objects and obj in self.objects[type(obj)]

    def __enter__(self):
        Network.context.append(self)
        self._config.__enter__()
        return self

    def __exit__(self, dummy_exc_type, dummy_exc_value, dummy_tb):
        if len(Network.context) == 0:
            raise NetworkContextError(
                "Network.context in bad state; was empty when "
                "exiting from a 'with' block."
            )

        config = Config.context[-1]
        if config is not self._config:
            raise ConfigError(
                "Config.context in bad state; was expecting "
                f"current context to be {self._config!r} but instead got "
                f"{config!r}."
            )

        network = Network.context.pop()
        if network is not self:
            raise NetworkContextError(
                "Network.context in bad state; was expecting current context "
                f"to be '{self}' but instead got '{network}'."
            )

        self._config.__exit__(dummy_exc_type, dummy_exc_value, dummy_tb)

    def __getstate__(self):
        state = self.__dict__.copy()
        state["label"] = self.label
        state["seed"] = self.seed
        return state

    def __setstate__(self, state):
        for k, v in state.items():
            setattr(self, k, v)
        if len(Network.context) > 0:
            warnings.warn(NotAddedToNetworkWarning(self))

    def __str__(self):
        label_txt = (
            f'"{self.label}"'
            if self.label is not None
            else f"(unlabeled) at 0x{id(self):x}"
        )
        return f"<{type(self).__name__} {label_txt}>"

    def __repr__(self):
        label_txt = f'"{self.label}"' if self.label is not None else "(unlabeled)"
        return f"<{type(self).__name__} {label_txt} at 0x{id(self):x}>"

    def copy(self, add_to_container=None):
        with warnings.catch_warnings():
            # We warn when copying since we can't change add_to_container.
            # However, we deal with it here, so we ignore the warning.
            warnings.simplefilter("ignore", category=NotAddedToNetworkWarning)
            c = deepcopy(self)
        if add_to_container is None:
            add_to_container = len(Network.context) > 0
        if add_to_container:
            Network.add(c)
        return c



import numpy as np
import pytest

import nengo
from nengo._vendor.npconv2d import conv2d
from nengo.exceptions import BuildError, ValidationError


@pytest.mark.parametrize("x_mul", (1, 2, 3, 4))
@pytest.mark.parametrize("k_size", (1, 2, 3, 4))
@pytest.mark.parametrize("stride", (1, 2, 3, 4))
@pytest.mark.parametrize("padding", ("same", "valid"))
def test_convolution_shape(padding, stride, k_size, x_mul, rng, allclose):
    tf = pytest.importorskip("tensorflow")

    in_channels = 2
    out_channels = 3

    for i in range(2 * k_size):
        x_size = k_size + stride * (x_mul - 1) + i
        x_shape = (x_size, x_size, in_channels)
        k_shape = (k_size, k_size, in_channels, out_channels)

        x = rng.uniform(-1, 1, size=x_shape)
        kernel = rng.uniform(-1, 1, size=k_shape)
        y_tf = tf.nn.conv2d(
            x[None, ...], kernel, stride, padding=padding.upper()
        ).numpy()[0]

        y_np = conv2d.conv2d(
            x[None, ...], kernel, pad=padding.upper(), stride=(stride, stride)
        )[0]

        transform = nengo.Convolution(
            out_channels,
            x_shape,
            kernel_size=(k_size, k_size),
            strides=(stride, stride),
            padding=padding,
        )

        assert transform.output_shape.shape == y_tf.shape
        assert y_np.shape == y_tf.shape
        assert allclose(y_np, y_tf)


@pytest.mark.parametrize("dimensions", (1, 2))
@pytest.mark.parametrize("padding", ("same", "valid"))
@pytest.mark.parametrize("channels_last", (True, False))
@pytest.mark.parametrize("fixed_kernel", (True, False))
@pytest.mark.parametrize("transpose", (True, False))
def test_convolution(
    dimensions,
    padding,
    channels_last,
    fixed_kernel,
    transpose,
    Simulator,
    allclose,
    rng,
    seed,
):
    input_d = 4
    input_channels = 2
    output_channels = 5
    kernel_d = 3
    output_d = (
        input_d
        if padding == "same"
        else input_d + (1 if transpose else -1) * (kernel_d // 2 * 2)
    )

    kernel_size = (kernel_d,) if dimensions == 1 else (kernel_d, kernel_d)
    input_shape = (input_d, input_channels)
    kernel_shape = (kernel_d, input_channels, output_channels)
    output_shape = (output_d, output_channels)

    if dimensions == 2:
        input_shape = (input_d,) + input_shape
        kernel_shape = (kernel_d,) + kernel_shape
        output_shape = (output_d,) + output_shape

    if not channels_last:
        input_shape = tuple(np.roll(input_shape, 1))
        output_shape = tuple(np.roll(output_shape, 1))

    x = rng.randn(*input_shape)
    w = rng.randn(*kernel_shape) if fixed_kernel else nengo.dists.Uniform(-0.1, 0.1)

    if transpose:
        transform = nengo.transforms.ConvolutionTranspose(
            output_channels,
            input_shape,
            init=w,
            padding=padding,
            kernel_size=kernel_size,
            strides=(1,) if dimensions == 1 else (1, 1),
            channels_last=channels_last,
        )
    else:
        transform = nengo.Convolution(
            output_channels,
            input_shape,
            init=w,
            padding=padding,
            kernel_size=kernel_size,
            strides=(1,) if dimensions == 1 else (1, 1),
            channels_last=channels_last,
        )

    assert transform.output_shape.shape == output_shape

    with nengo.Network(seed=seed) as net:
        a = nengo.Node(np.ravel(x))
        b = nengo.Node(size_in=np.prod(output_shape))
        conn = nengo.Connection(a, b, synapse=None, transform=transform)
        p = nengo.Probe(b)

        # check error handling
        bad_in = nengo.Node([0])
        bad_out = nengo.Node(size_in=5)
        with pytest.raises(ValidationError):
            nengo.Connection(bad_in, b, transform=conn.transform)
        with pytest.raises(ValidationError):
            nengo.Connection(a, bad_out, transform=conn.transform)

    assert conn.transform.output_shape.shape == output_shape
    assert conn.transform.kernel_shape == kernel_shape

    with Simulator(net) as sim:
        sim.step()

    weights = sim.data[conn].weights
    if not channels_last:
        x = np.moveaxis(x, 0, -1)
    if dimensions == 1:
        x = x[:, None, :]
        weights = weights[:, None, :, :]

    if transpose:
        outsize = (output_d, 1) if dimensions == 1 else (output_d, output_d)
        truth = conv2d.conv2d_gradx(
            weights, x[None, ...], xsize=outsize, pad=padding.upper()
        )[0]
    else:
        truth = conv2d.conv2d(x[None, ...], weights, pad=padding.upper())[0]

    if not channels_last:
        truth = np.moveaxis(truth, -1, 0)

    assert allclose(sim.data[p][0], np.ravel(truth))


@pytest.mark.parametrize("encoders", (True, False))
@pytest.mark.parametrize("decoders", (True, False))
def test_convolution_nef(encoders, decoders, Simulator):
    with nengo.Network() as net:
        transform = nengo.transforms.Convolution(n_filters=2, input_shape=(3, 3, 1))
        a = nengo.Ensemble(9, 9)
        b = nengo.Ensemble(2, 2)
        nengo.Connection(
            a if decoders else a.neurons,
            b if encoders else b.neurons,
            transform=transform,
        )

    if decoders:
        # error if decoders
        with pytest.raises(BuildError, match="decoded connection"):
            with Simulator(net):
                pass
    else:
        # no error
        with Simulator(net):
            pass


def test_convolution_validation_errors():
    # conflicting channels_last
    input_shape = nengo.transforms.ChannelShape((2, 3, 4), channels_last=True)
    with pytest.raises(ValidationError, match="transform has channels_l.*input shape"):
        nengo.Convolution(4, input_shape, channels_last=False)

    # kernel_size does not match dimensions (2)
    with pytest.raises(ValidationError, match=r"Kernel dimensions \(3\) does not mat"):
        nengo.Convolution(4, input_shape, kernel_size=(3, 3, 3))

    # strides does not match dimensions (2)
    with pytest.raises(ValidationError, match=r"Stride dimensions \(3\) does not mat"):
        nengo.Convolution(4, input_shape, strides=(1, 1, 1))

    # init shape does not match kernel shape
    input_shape = nengo.transforms.ChannelShape((5, 5, 4), channels_last=True)
    nengo.Convolution(4, input_shape, init=np.ones((3, 3, 4, 4)))  # this works
    with pytest.raises(ValidationError, match=r"Kernel shape \(9, 9, 4, 4\).*not mat"):
        nengo.Convolution(4, input_shape, init=np.ones((9, 9, 4, 4)))
    with pytest.raises(ValidationError, match=r"Kernel shape \(3, 3, 7, 4\).*not mat"):
        nengo.Convolution(4, input_shape, init=np.ones((3, 3, 7, 4)))
    with pytest.raises(ValidationError, match=r"Kernel shape \(3, 3, 4, 5\).*not mat"):
        nengo.Convolution(4, input_shape, init=np.ones((3, 3, 4, 5)))

    # test empty output
    with pytest.raises(ValidationError, match="exceeds the spatial size"):
        nengo.transforms.Convolution(n_filters=2, input_shape=(3, 2, 1))

    # valid output shape
    nengo.transforms.ConvolutionTranspose(
        n_filters=2, input_shape=(3, 2, 1), output_shape=(5, 4, 2)
    )
    with pytest.raises(ValidationError, match="number of dimensions"):
        # too many dims in output shape
        nengo.transforms.ConvolutionTranspose(
            n_filters=2, input_shape=(3, 2, 1), output_shape=(5, 4, 2, 1)
        )
    with pytest.raises(ValidationError, match="number of channels"):
        # too many channels in output shape
        nengo.transforms.ConvolutionTranspose(
            n_filters=2, input_shape=(3, 2, 1), output_shape=(5, 4, 3)
        )
    with pytest.raises(ValidationError, match="not a valid output shape"):
        # too many rows in output shape
        nengo.transforms.ConvolutionTranspose(
            n_filters=2, input_shape=(3, 2, 1), output_shape=(6, 4, 2)
        )



"""Reference simulator for nengo models.

This backend is relatively fast, and works on general purpose computers.
Other Nengo backends provide more specialized Simulators for custom platforms.
"""

import logging
import warnings
from collections.abc import Mapping

import numpy as np

import nengo.utils.numpy as npext
from nengo.builder import Model
from nengo.builder.optimizer import optimize as opmerge_optimize
from nengo.builder.signal import SignalDict
from nengo.cache import get_default_decoder_cache
from nengo.exceptions import ReadonlyError, SimulatorClosed, ValidationError
from nengo.utils.graphs import toposort
from nengo.utils.progress import Progress, ProgressTracker
from nengo.utils.simulator import operator_dependency_graph

logger = logging.getLogger(__name__)


class Simulator:
    r"""Reference simulator for Nengo models.

    The simulator takes a `.Network` and builds internal data structures to
    run the model defined by that network. Run the simulator with the
    `~.Simulator.run` method, and access probed data through the
    ``data`` attribute.

    Building and running the simulation may allocate resources like files
    and sockets. To properly free these resources, call the `.Simulator.close`
    method. Alternatively, `.Simulator.close` will automatically be called
    if you use the ``with`` syntax:

    .. testcode::

       with nengo.Network() as net:
           ensemble = nengo.Ensemble(10, 1)
       with nengo.Simulator(net, progress_bar=False) as sim:
           sim.run(0.1)

    Note that the ``data`` attribute is still accessible even when a simulator
    has been closed. Running the simulator, however, will raise an error.

    When debugging or comparing models, it can be helpful to see the full ordered
    list of operators that the simulator will execute on each timestep.

    .. testcode::

       with nengo.Simulator(nengo.Network(), progress_bar=False) as sim:
           print('\n'.join(f"* {op}" for op in sim.step_order))

    .. testoutput::

       * TimeUpdate{}

    The diff of two simulators' sorted ops tells us how two built models differ.
    We can use ``difflib`` in the Python standard library to see the differences.

    .. testcode::

       # Original model
       with nengo.Network() as net:
           ensemble = nengo.Ensemble(10, 1, label="Ensemble")
       sim1 = nengo.Simulator(net, progress_bar=False)

       # Add a node
       with net:
           node = nengo.Node(output=0, label="Node")
           nengo.Connection(node, ensemble)
       sim2 = nengo.Simulator(net, progress_bar=False)

       import difflib

       print("".join(difflib.unified_diff(
           sorted(f"{type(op).__name__}: {op.tag}\n" for op in sim1.step_order),
           sorted(f"{type(op).__name__}: {op.tag}\n" for op in sim2.step_order),
           fromfile="sim1",
           tofile="sim2",
           n=0,
       )).strip())

       sim1.close()
       sim2.close()

    .. testoutput::
       :options:

       --- sim1
       +++ sim2
       @@ -0,0 +1 @@
       +Copy: <Connection from <Node 'Node'> to <Ensemble 'Ensemble'>>
       @@ -4,0 +6 @@
       +SimProcess: Lowpass(tau=0.005)

    Parameters
    ----------
    network : Network or None
        A network object to be built and then simulated. If None,
        then a `.Model` with the build model must be provided instead.
    dt : float, optional
        The length of a simulator timestep, in seconds.
    seed : int, optional
        A seed for all stochastic operators used in this simulator.
        Will be set to ``network.seed + 1`` if not given.
    model : Model, optional
        A `.Model` that contains build artifacts to be simulated.
        Usually the simulator will build this model for you; however, if you
        want to build the network manually, or you want to inject build
        artifacts in the model before building the network, then you can
        pass in a `.Model` instance.
    progress_bar : bool or ProgressBar, optional
        Progress bar for displaying build and simulation progress.

        If ``True``, the default progress bar will be used.
        If ``False``, the progress bar will be disabled.
        For more control over the progress bar, pass in a ``ProgressBar``
        instance.
    optimize : bool, optional
        If ``True``, the builder will run an additional optimization step
        that can speed up simulations significantly at the cost of slower
        builds. If running models for very small amounts of time,
        pass ``False`` to disable the optimizer.

    Attributes
    ----------
    closed : bool
        Whether the simulator has been closed.
        Once closed, it cannot be reopened.
    data : SimulationData
        The `.SimulationData` mapping from Nengo objects to the data associated
        with those objects. In particular, each `.Probe` maps to the data
        probed while running the simulation.
    dg : dict
        A dependency graph mapping from each `.Operator` to the operators
        that depend on that operator.
    model : Model
        The `.Model` containing the signals and operators necessary to
        simulate the network.
    signals : SignalDict
        The `.SignalDict` mapping from `.Signal` instances to NumPy arrays.

    """

    def __init__(
        self, network, dt=0.001, seed=None, model=None, progress_bar=True, optimize=True
    ):
        self.closed = True  # Start closed in case constructor raises exception
        self.progress_bar = progress_bar
        self.optimize = optimize

        if model is None:
            self.model = Model(
                dt=float(dt),
                label=f"{network}, dt={dt:f}",
                decoder_cache=get_default_decoder_cache(),
            )
        else:
            self.model = model

        pt = ProgressTracker(progress_bar, Progress("Building", "Build"))
        with pt:
            if network is not None:
                # Build the network into the model
                self.model.build(network, progress=pt.next_stage("Building", "Build"))

            # Order the steps (they are made in `Simulator.reset`)
            self.dg = operator_dependency_graph(self.model.operators)

            if optimize:
                with pt.next_stage("Building (running optimizer)", "Optimization"):
                    opmerge_optimize(self.model, self.dg)

        self._step_order = [op for op in toposort(self.dg) if hasattr(op, "make_step")]

        # -- map from Signal.base -> ndarray
        self.signals = SignalDict()
        for op in self.model.operators:
            op.init_signals(self.signals)

        # Add built states to the raw simulation data dictionary
        self._sim_data = self.model.params

        # Provide a nicer interface to simulation data
        self.data = SimulationData(self._sim_data)

        if seed is None:
            if network is not None and network.seed is not None:
                seed = network.seed + 1
            else:
                seed = np.random.randint(npext.maxint)

        self.closed = False
        self.reset(seed=seed)

    def __del__(self):
        """Raise a ResourceWarning if we are deallocated while open."""
        if not self.closed:
            warnings.warn(
                f"Simulator with model={self.model} was deallocated while open. Please "
                "close simulators manually to ensure resources are properly "
                "freed.",
                ResourceWarning,
            )

    def __enter__(self):
        if self.closed:
            raise SimulatorClosed("Cannot re-open after simulator is closed")

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def __getstate__(self):
        signals = (
            {k: v for k, v in self.signals.items() if not k.readonly}
            if self.signals is not None
            else {}
        )
        probe_outputs = {probe: self._sim_data[probe] for probe in self.model.probes}
        return dict(
            model=self.model,
            signals=signals,
            probe_outputs=probe_outputs,
            dt=self.dt,
            seed=self.seed,
            progress_bar=self.progress_bar,
            optimize=self.optimize,
            closed=self.closed,
        )

    def __setstate__(self, state):
        self.__init__(
            network=None,
            model=state["model"],
            dt=state["dt"],
            seed=state["seed"],
            progress_bar=state["progress_bar"],
            optimize=False,  # The pickled Sim will have already been optimized
        )
        for key, value in state["signals"].items():
            self.signals[key] = value
        for key, value in state["probe_outputs"].items():
            self._sim_data[key].extend(value)
        # Set whether it had originally been optimized
        self.optimize = state["optimize"]
        if state["closed"]:
            self.close()

    @property
    def dt(self):
        """(float) The time step of the simulator."""
        return self.model.dt

    @dt.setter
    def dt(self, dummy):
        raise ReadonlyError(attr="dt", obj=self)

    @property
    def n_steps(self):
        """(int) The current time step of the simulator."""
        return self._n_steps

    @property
    def step_order(self):
        """(list) The ordered list of step functions run by this simulator."""
        return self._step_order

    @property
    def time(self):
        """(float) The current time of the simulator."""
        return self._time

    def clear_probes(self):
        """Clear all probe histories.

        .. versionadded:: 3.0.0
        """
        for probe in self.model.probes:
            self._sim_data[probe] = []
        self.data.reset()  # clear probe cache

    def close(self):
        """Closes the simulator.

        Any call to `.Simulator.run`, `.Simulator.run_steps`,
        `.Simulator.step`, and `.Simulator.reset` on a closed simulator raises
        a `.SimulatorClosed` exception.
        """
        self.closed = True
        self.signals = None  # signals may no longer exist on some backends

    def _probe(self):
        """Copy all probed signals to buffers."""
        self._probe_step_time()

        for probe in self.model.probes:
            period = 1 if probe.sample_every is None else probe.sample_every / self.dt
            if self.n_steps % period < 1:
                tmp = self.signals[self.model.sig[probe]["in"]].copy()
                self._sim_data[probe].append(tmp)

    def _probe_step_time(self):
        self._n_steps = self.signals[self.model.step].item()
        self._time = self.signals[self.model.time].item()

    def reset(self, seed=None):
        """Reset the simulator state.

        Parameters
        ----------
        seed : int, optional
            A seed for all stochastic operators used in the simulator.
            This will change the random sequences generated for noise
            or inputs (e.g. from processes), but not the built objects
            (e.g. ensembles, connections).
        """
        if self.closed:
            raise SimulatorClosed("Cannot reset closed Simulator.")

        if seed is not None:
            self.seed = seed

        # reset signals
        for key in self.signals:
            self.signals.reset(key)

        # rebuild steps (resets ops with their own state, like Processes)
        self.rng = np.random.RandomState(self.seed)
        self._steps = [
            op.make_step(self.signals, self.dt, self.rng) for op in self._step_order
        ]

        self.clear_probes()

        self._probe_step_time()

    def run(self, time_in_seconds, progress_bar=None):
        """Simulate for the given length of time.

        If the given length of time is not a multiple of ``dt``,
        it will be rounded to the nearest ``dt``. For example, if ``dt``
        is 0.001 and ``run`` is called with ``time_in_seconds=0.0006``,
        the simulator will advance one timestep, resulting in the actual
        simulator time being 0.001.

        The given length of time must be positive. The simulator cannot
        be run backwards.

        Parameters
        ----------
        time_in_seconds : float
            Amount of time to run the simulation for. Must be positive.
        progress_bar : bool or ProgressBar, optional
            Progress bar for displaying the progress of the simulation run.

            If True, the default progress bar will be used.
            If False, the progress bar will be disabled.
            For more control over the progress bar, pass in a ``ProgressBar``
            instance.
        """
        if time_in_seconds < 0:
            raise ValidationError(
                f"Must be positive (got {time_in_seconds:g})", attr="time_in_seconds"
            )

        steps = int(np.round(float(time_in_seconds) / self.dt))

        if steps == 0:
            warnings.warn(
                f"{time_in_seconds} results in running for 0 timesteps. Simulator "
                f"still at time {self.time}."
            )
        else:
            logger.info(
                "Running %s for %f seconds, or %d steps",
                self.model.label,
                time_in_seconds,
                steps,
            )
            self.run_steps(steps, progress_bar=progress_bar)

    def run_steps(self, steps, progress_bar=None):
        """Simulate for the given number of ``dt`` steps.

        Parameters
        ----------
        steps : int
            Number of steps to run the simulation for.
        progress_bar : bool or ProgressBar, optional
            Progress bar for displaying the progress of the simulation run.

            If True, the default progress bar will be used.
            If False, the progress bar will be disabled.
            For more control over the progress bar, pass in a ``ProgressBar``
            instance.
        """
        if progress_bar is None:
            progress_bar = self.progress_bar

        with ProgressTracker(
            progress_bar, Progress("Simulating", "Simulation", steps)
        ) as pt:
            for _ in range(steps):
                self.step()
                pt.total_progress.step()

    def step(self):
        """Advance the simulator by 1 step (``dt`` seconds)."""
        if self.closed:
            raise SimulatorClosed("Simulator cannot run because it is closed.")

        old_err = np.seterr(invalid="raise", divide="ignore")
        try:
            for step_fn in self._steps:
                step_fn()
        finally:
            np.seterr(**old_err)

        self._probe()

    def trange(self, dt=None, sample_every=None):
        """Create a vector of times matching probed data.

        Note that the range does not start at 0 as one might expect, but at
        the first timestep (i.e., ``dt``).

        Parameters
        ----------
        sample_every : float, optional
            The sampling period of the probe to create a range for.
            If None, a time value for every ``dt`` will be produced.

            .. versionchanged:: 3.0.0
               Renamed from dt to sample_every
        """
        if dt is not None:
            if sample_every is not None:
                raise ValidationError(
                    "Cannot specify both `dt` and `sample_every`. "
                    "Use `sample_every` only.",
                    attr="dt",
                    obj=self,
                )
            warnings.warn(
                "`dt` is deprecated. Use `sample_every` instead.", DeprecationWarning
            )
            sample_every = dt
        period = 1 if sample_every is None else sample_every / self.dt
        steps = np.arange(1, self.n_steps + 1)
        return self.dt * steps[steps % period < 1]


class SimulationData(Mapping):
    """Data structure used to access simulation data from the model.

    The main use case for this is to access Probe data; for example,
    ``probe_data = sim.data[my_probe]``. However, it is also used to access the
    parameters of objects in the model; for example, encoder values for an ensemble
    can be accessed via ``encoders = sim.data[my_ens].encoders``.

    This is like a view on the raw simulation data manipulated by the Simulator,
    which allows the raw simulation data to be optimized for speed while this
    provides a more user-friendly interface.

    .. versionchanged:: 3.0.0
       Renamed from ProbeDict to SimulationData
    """

    def __init__(self, raw):
        super().__init__()
        self.raw = raw
        self._cache = {}

    def __getitem__(self, key):
        """Return simulation data for ``key`` object.

        For speed reasons, the simulator uses Python lists for Probe data
        and we want to return NumPy arrays.
        """
        if key not in self._cache or len(self._cache[key]) != len(self.raw[key]):
            rval = self.raw[key]
            if isinstance(rval, list):
                rval = np.asarray(rval)
                rval.setflags(write=False)
            self._cache[key] = rval
        return self._cache[key]

    def __iter__(self):
        return iter(self.raw)

    def __len__(self):
        return len(self.raw)

    def __repr__(self):
        return repr(self.raw)

    def __str__(self):
        return str(self.raw)

    def reset(self):
        self._cache.clear()



import pytest

import nengo
from nengo.exceptions import ObsoleteError


def test_net_arg_obsoleteerror():
    n_neurons = 10
    dimensions = 1
    with nengo.Network() as net:
        with pytest.raises(ObsoleteError):
            nengo.networks.CircularConvolution(n_neurons, dimensions, net=net)
        with pytest.raises(ObsoleteError):
            nengo.networks.Integrator(0.1, n_neurons, dimensions, net=net)
        with pytest.raises(ObsoleteError):
            nengo.networks.Oscillator(0.1, 1, n_neurons, net=net)
        with pytest.raises(ObsoleteError):
            nengo.networks.Product(n_neurons, dimensions, net=net)
        with pytest.raises(ObsoleteError):
            nengo.networks.InputGatedMemory(n_neurons, dimensions, net=net)

        with pytest.raises(ObsoleteError):
            nengo.networks.BasalGanglia(dimensions, net=net)
        with pytest.raises(ObsoleteError):
            nengo.networks.Thalamus(dimensions, net=net)



import warnings

from nengo.config import SupportDefaultsMixin
from nengo.exceptions import ValidationError
from nengo.params import Default, FrozenObject, IntParam, NumberParam, Parameter
from nengo.synapses import Lowpass, SynapseParam
from nengo.utils.numpy import is_iterable


class LearningRuleTypeSizeInParam(IntParam):
    valid_strings = ("pre", "post", "mid", "pre_state", "post_state")

    def coerce(self, instance, size_in):  # pylint: disable=arguments-renamed
        if isinstance(size_in, str):
            if size_in not in self.valid_strings:
                raise ValidationError(
                    f"{size_in!r} is not a valid string value "
                    f"(must be one of {self.valid_strings})",
                    attr=self.name,
                    obj=instance,
                )
            return size_in
        else:
            return super().coerce(instance, size_in)  # IntParam validation


class LearningRuleType(FrozenObject, SupportDefaultsMixin):
    """Base class for all learning rule objects.

    To use a learning rule, pass it as a ``learning_rule_type`` keyword
    argument to the `~nengo.Connection` on which you want to do learning.

    Each learning rule exposes two important pieces of metadata that the
    builder uses to determine what information should be stored.

    The ``size_in`` is the dimensionality of the incoming error signal. It
    can either take an integer or one of the following string values:

    * ``'pre'``: vector error signal in pre-object space
    * ``'post'``: vector error signal in post-object space
    * ``'mid'``: vector error signal in the ``conn.size_mid`` space
    * ``'pre_state'``: vector error signal in pre-synaptic ensemble space
    * ``'post_state'``: vector error signal in post-synaptic ensemble space

    The difference between ``'post_state'`` and ``'post'`` is that with the
    former, if a ``Neurons`` object is passed, it will use the dimensionality
    of the corresponding ``Ensemble``, whereas the latter simply uses the
    ``post`` object ``size_in``. Similarly with ``'pre_state'`` and ``'pre'``.

    The ``modifies`` attribute denotes the signal targeted by the rule.
    Options are:

    * ``'encoders'``
    * ``'decoders'``
    * ``'weights'``

    Parameters
    ----------
    learning_rate : float, optional
        A scalar indicating the rate at which ``modifies`` will be adjusted.
    size_in : int, str, optional
        Dimensionality of the error signal (see above).

    Attributes
    ----------
    learning_rate : float
        A scalar indicating the rate at which ``modifies`` will be adjusted.
    size_in : int, str
        Dimensionality of the error signal.
    modifies : str
        The signal targeted by the learning rule.
    """

    modifies = None
    probeable = ()

    learning_rate = NumberParam("learning_rate", low=0, readonly=True, default=1e-6)
    size_in = LearningRuleTypeSizeInParam("size_in", low=0)

    def __init__(self, learning_rate=Default, size_in=0):
        super().__init__()
        self.learning_rate = learning_rate
        self.size_in = size_in


class PES(LearningRuleType):
    """Prescribed Error Sensitivity learning rule.

    Modifies a connection's decoders to minimize an error signal provided
    through a connection to the connection's learning rule.

    Parameters
    ----------
    learning_rate : float, optional
        A scalar indicating the rate at which weights will be adjusted.
    pre_synapse : `.Synapse`, optional
        Synapse model used to filter the pre-synaptic activities.

    Attributes
    ----------
    learning_rate : float
        A scalar indicating the rate at which weights will be adjusted.
    pre_synapse : `.Synapse`
        Synapse model used to filter the pre-synaptic activities.
    """

    modifies = "decoders"
    probeable = ("error", "activities", "delta")

    learning_rate = NumberParam("learning_rate", low=0, readonly=True, default=1e-4)
    pre_synapse = SynapseParam("pre_synapse", default=Lowpass(tau=0.005), readonly=True)

    def __init__(self, learning_rate=Default, pre_synapse=Default):
        super().__init__(learning_rate, size_in="post_state")
        if learning_rate is not Default and learning_rate >= 1.0:
            warnings.warn(
                "This learning rate is very high, and can result "
                "in floating point errors from too much current."
            )

        self.pre_synapse = pre_synapse


class RLS(LearningRuleType):
    r"""Recursive least-squares rule for online decoder optimization.

    This implements an online version of the standard least-squares solvers used
    to learn connection weights offline (e.g. `nengo.solvers.LstsqL2`). It can be
    applied in the same scenarios as `.PES`, to minimize an error signal.

    The cost of RLS is :math:`\mathcal{O}(n^2)` extra time and memory. If possible,
    it is more efficient to do the learning offline using e.g. `.LstsqL2`.

    Parameters
    ----------
    learning_rate : float, optional
        Effective learning rate. This is better understood as
        :math:`\frac{1}{\alpha}`, where :math:`\alpha` is an L2-regularization
        term. A large learning rate means little regularization, which implies
        quick over-fitting. A small learning rate means large regularization,
        which translates to slower learning [#]_.
    pre_synapse : Synapse, optional
        Synapse model applied to the pre-synaptic neural activities.

    See Also
    --------
    nengo.PES
    nengo.solvers.LstsqL2

    Notes
    -----
    RLS works by maintaining the inverse neural correlation matrix,
    :math:`P = \Gamma^{-1}`, where :math:`\Gamma = A^T A + \alpha I` are the
    regularized correlations, :math:`A` is a matrix of (possibly filtered)
    neural activities, and :math:`\alpha` is an L2-regularization term
    controlled by the ``learning_rate``. :math:`P` is used to project the
    error signal and update the weights each time-step.

    References
    ----------
    .. [#] Sussillo, D., & Abbott, L. F. (2009). Generating coherent patterns
       of activity from chaotic neural networks. Neuron, 63(4), 544-557.

    Examples
    --------
    Below, we compare `.PES` against `.RLS`, learning a feed-forward
    communication channel (identity function) online, starting with 100 spiking
    LIF neurons with decoders (weights) set to zero. A faster learning rate for
    `.PES` results in over-fitting to the most recent online example, while a
    slower learning rate does not learn quickly enough. This is a general problem
    with greedy optimization. `.RLS` performs better since it is L2-optimal.

    .. testcode::

       from nengo.learning_rules import PES, RLS

       tau = 0.005
       learning_rules = (
           PES(learning_rate=1e-3, pre_synapse=tau),
           RLS(learning_rate=1e-3, pre_synapse=tau),
       )

       with nengo.Network() as model:
           u = nengo.Node(output=lambda t: np.sin(2 * np.pi * t))
           probes = []
           for lr in learning_rules:
               e = nengo.Node(size_in=1, output=lambda t, e: e if t < 1 else 0)
               x = nengo.Ensemble(100, 1, seed=0)
               y = nengo.Node(size_in=1)

               nengo.Connection(u, e, synapse=None, transform=-1)
               nengo.Connection(u, x, synapse=None)
               conn = nengo.Connection(
                   x, y, synapse=None, learning_rule_type=lr, function=lambda x: 0
               )
               nengo.Connection(y, e, synapse=None)
               nengo.Connection(e, conn.learning_rule, synapse=tau)
               probes.append(nengo.Probe(y, synapse=tau))
           probes.append(nengo.Probe(u, synapse=tau))

       with nengo.Simulator(model) as sim:
           sim.run(2.0)

       plt.plot(sim.trange(), sim.data[probes[0]], label=str(learning_rules[0]))
       plt.plot(sim.trange(), sim.data[probes[1]], label=str(learning_rules[1]))
       plt.plot(sim.trange(), sim.data[probes[2]], label="Ideal", linestyle="--")
       plt.vlines([1], -1, 1, label="Training -> Testing")
       plt.ylim(-2, 2)
       plt.legend(loc="upper right")
       plt.xlabel("Time (s)")

    .. testoutput::
      :hide:

      ...
    """

    modifies = "decoders"
    probeable = ("pre_filtered", "error", "delta", "inv_gamma")

    learning_rate = NumberParam("learning_rate", low=0, readonly=True, default=1e-3)
    pre_synapse = SynapseParam("pre_synapse", default=Lowpass(tau=0.005), readonly=True)

    def __init__(self, learning_rate=Default, pre_synapse=Default):
        super().__init__(learning_rate=learning_rate, size_in="post_state")
        self.pre_synapse = pre_synapse


def _remove_default_post_synapse(argreprs, default):
    default_post_synapse = f"post_synapse={default!r}"
    if default_post_synapse in argreprs:
        argreprs.remove(default_post_synapse)
    return argreprs


class BCM(LearningRuleType):
    """Bienenstock-Cooper-Munroe learning rule.

    Modifies connection weights as a function of the presynaptic activity
    and the difference between the postsynaptic activity and the average
    postsynaptic activity.

    Notes
    -----
    The BCM rule is dependent on pre and post neural activities,
    not decoded values, and so is not affected by changes in the
    size of pre and post ensembles. However, if you are decoding from
    the post ensemble, the BCM rule will have an increased effect on
    larger post ensembles because more connection weights are changing.
    In these cases, it may be advantageous to scale the learning rate
    on the BCM rule by ``1 / post.n_neurons``.

    Parameters
    ----------
    learning_rate : float, optional
        A scalar indicating the rate at which weights will be adjusted.
    pre_synapse : `.Synapse`, optional
        Synapse model used to filter the pre-synaptic activities.
    post_synapse : `.Synapse`, optional
        Synapse model used to filter the post-synaptic activities.
        If None, ``post_synapse`` will be the same as ``pre_synapse``.
    theta_synapse : `.Synapse`, optional
        Synapse model used to filter the theta signal.

    Attributes
    ----------
    learning_rate : float
        A scalar indicating the rate at which weights will be adjusted.
    post_synapse : `.Synapse`
        Synapse model used to filter the post-synaptic activities.
    pre_synapse : `.Synapse`
        Synapse model used to filter the pre-synaptic activities.
    theta_synapse : `.Synapse`
        Synapse model used to filter the theta signal.
    """

    modifies = "weights"
    probeable = ("theta", "pre_filtered", "post_filtered", "delta")

    learning_rate = NumberParam("learning_rate", low=0, readonly=True, default=1e-9)
    pre_synapse = SynapseParam("pre_synapse", default=Lowpass(tau=0.005), readonly=True)
    post_synapse = SynapseParam("post_synapse", default=None, readonly=True)
    theta_synapse = SynapseParam(
        "theta_synapse", default=Lowpass(tau=1.0), readonly=True
    )

    def __init__(
        self,
        learning_rate=Default,
        pre_synapse=Default,
        post_synapse=Default,
        theta_synapse=Default,
    ):
        super().__init__(learning_rate, size_in=0)

        self.pre_synapse = pre_synapse
        self.post_synapse = (
            self.pre_synapse if post_synapse is Default else post_synapse
        )
        self.theta_synapse = theta_synapse

    @property
    def _argreprs(self):
        return _remove_default_post_synapse(super()._argreprs, self.pre_synapse)


class Oja(LearningRuleType):
    """Oja learning rule.

    Modifies connection weights according to the Hebbian Oja rule, which
    augments typically Hebbian coactivity with a "forgetting" term that is
    proportional to the weight of the connection and the square of the
    postsynaptic activity.

    Notes
    -----
    The Oja rule is dependent on pre and post neural activities,
    not decoded values, and so is not affected by changes in the
    size of pre and post ensembles. However, if you are decoding from
    the post ensemble, the Oja rule will have an increased effect on
    larger post ensembles because more connection weights are changing.
    In these cases, it may be advantageous to scale the learning rate
    on the Oja rule by ``1 / post.n_neurons``.

    Parameters
    ----------
    learning_rate : float, optional
        A scalar indicating the rate at which weights will be adjusted.
    pre_synapse : `.Synapse`, optional
        Synapse model used to filter the pre-synaptic activities.
    post_synapse : `.Synapse`, optional
        Synapse model used to filter the post-synaptic activities.
        If None, ``post_synapse`` will be the same as ``pre_synapse``.
    beta : float, optional
        A scalar weight on the forgetting term.

    Attributes
    ----------
    beta : float
        A scalar weight on the forgetting term.
    learning_rate : float
        A scalar indicating the rate at which weights will be adjusted.
    post_synapse : `.Synapse`
        Synapse model used to filter the post-synaptic activities.
    pre_synapse : `.Synapse`
        Synapse model used to filter the pre-synaptic activities.
    """

    modifies = "weights"
    probeable = ("pre_filtered", "post_filtered", "delta")

    learning_rate = NumberParam("learning_rate", low=0, readonly=True, default=1e-6)
    pre_synapse = SynapseParam("pre_synapse", default=Lowpass(tau=0.005), readonly=True)
    post_synapse = SynapseParam("post_synapse", default=None, readonly=True)
    beta = NumberParam("beta", low=0, readonly=True, default=1.0)

    def __init__(
        self,
        learning_rate=Default,
        pre_synapse=Default,
        post_synapse=Default,
        beta=Default,
    ):
        super().__init__(learning_rate, size_in=0)

        self.beta = beta
        self.pre_synapse = pre_synapse
        self.post_synapse = (
            self.pre_synapse if post_synapse is Default else post_synapse
        )

    @property
    def _argreprs(self):
        return _remove_default_post_synapse(super()._argreprs, self.pre_synapse)


class Voja(LearningRuleType):
    """Vector Oja learning rule.

    Modifies an ensemble's encoders to be selective to its inputs.

    A connection to the learning rule will provide a scalar weight for the
    learning rate, minus 1. For instance, 0 is normal learning, -1 is no
    learning, and less than -1 causes anti-learning or "forgetting".

    Parameters
    ----------
    post_tau : float, optional
        Filter constant on activities of neurons in post population.
    learning_rate : float, optional
        A scalar indicating the rate at which encoders will be adjusted.
    post_synapse : `.Synapse`, optional
        Synapse model used to filter the post-synaptic activities.

    Attributes
    ----------
    learning_rate : float
        A scalar indicating the rate at which encoders will be adjusted.
    post_synapse : `.Synapse`
        Synapse model used to filter the post-synaptic activities.
    """

    modifies = "encoders"
    probeable = ("post_filtered", "scaled_encoders", "delta")

    learning_rate = NumberParam("learning_rate", low=0, readonly=True, default=1e-2)
    post_synapse = SynapseParam(
        "post_synapse", default=Lowpass(tau=0.005), readonly=True
    )

    def __init__(self, learning_rate=Default, post_synapse=Default):
        super().__init__(learning_rate, size_in=1)

        self.post_synapse = post_synapse


class LearningRuleTypeParam(Parameter):
    def check_rule(self, instance, rule):
        if not isinstance(rule, LearningRuleType):
            raise ValidationError(
                f"'{rule}' must be a learning rule type or a dict or "
                "list of such types.",
                attr=self.name,
                obj=instance,
            )
        if rule.modifies not in ("encoders", "decoders", "weights"):
            raise ValidationError(
                f"Unrecognized target {rule.modifies!r}", attr=self.name, obj=instance
            )

    def coerce(self, instance, rule):  # pylint: disable=arguments-renamed
        if is_iterable(rule):
            for r in rule.values() if isinstance(rule, dict) else rule:
                self.check_rule(instance, r)
        elif rule is not None:
            self.check_rule(instance, rule)
        return super().coerce(instance, rule)



import numpy as np

from nengo.base import Process
from nengo.exceptions import ValidationError
from nengo.params import (
    BoolParam,
    EnumParam,
    NdarrayParam,
    NumberParam,
    Parameter,
    Unconfigurable,
)
from nengo.rc import rc
from nengo.utils.filter_design import cont2discrete, tf2ss
from nengo.utils.numpy import as_shape, is_number


class Synapse(Process):
    """Abstract base class for synapse models.

    Conceptually, a synapse model emulates a biological synapse, taking in
    input in the form of released neurotransmitter and opening ion channels
    to allow more or less current to flow into the neuron.

    In Nengo, the implementation of a synapse is as a specific case of a
    `.Process` in which the input and output shapes are the same.
    The input is the current across the synapse, and the output is the current
    that will be induced in the postsynaptic neuron.

    Synapses also contain the `.Synapse.filt` and `.Synapse.filtfilt` methods,
    which make it easy to use Nengo's synapse models outside of Nengo
    simulations.

    Parameters
    ----------
    default_size_in : int, optional
        The size_in used if not specified.
    default_size_out : int
        The size_out used if not specified.
        If None, will be the same as default_size_in.
    default_dt : float
        The simulation timestep used if not specified.
    seed : int, optional
        Random number seed. Ensures random factors will be the same each run.

    Attributes
    ----------
    default_dt : float
        The simulation timestep used if not specified.
    default_size_in : int
        The size_in used if not specified.
    default_size_out : int
        The size_out used if not specified.
    seed : int, optional
        Random number seed. Ensures random factors will be the same each run.
    """

    def __init__(
        self, default_size_in=1, default_size_out=None, default_dt=0.001, seed=None
    ):
        if default_size_out is None:
            default_size_out = default_size_in
        super().__init__(
            default_size_in=default_size_in,
            default_size_out=default_size_out,
            default_dt=default_dt,
            seed=seed,
        )

    def make_state(self, shape_in, shape_out, dt, dtype=None, y0=None):
        raise NotImplementedError("Synapse must implement make_state")

    def filt(self, x, dt=None, axis=0, y0=0, copy=True, filtfilt=False):
        """Filter ``x`` with this synapse model.

        Parameters
        ----------
        x : array_like
            The signal to filter.
        dt : float, optional
            The timestep of the input signal.
            If None, ``default_dt`` will be used.
        axis : int, optional
            The axis along which to filter.
        y0 : array_like, optional
            The starting state of the filter output. Must be zero for
            unstable linear systems.
        copy : bool, optional
            Whether to copy the input data, or simply work in-place.
        filtfilt : bool, optional
            If True, runs the process forward then backward on the signal,
            for zero-phase filtering (like Matlab's ``filtfilt``).
        """
        # This function is very similar to `Process.apply`, but allows for
        # a) filtering along any axis, and b) zero-phase filtering (filtfilt).
        dt = self.default_dt if dt is None else dt
        filtered = np.array(x, copy=copy, dtype=rc.float_dtype)
        filt_view = np.rollaxis(filtered, axis=axis)  # rolled view on filtered

        shape_in = shape_out = as_shape(filt_view[0].shape, min_dim=1)
        state = self.make_state(shape_in, shape_out, dt, dtype=filtered.dtype, y0=y0)
        step = self.make_step(shape_in, shape_out, dt, rng=None, state=state)

        for i, signal_in in enumerate(filt_view):
            filt_view[i] = step(i * dt, signal_in)

        if filtfilt:  # Flip the filt_view and filter again
            n = len(filt_view) - 1
            filt_view = filt_view[::-1]
            for i, signal_in in enumerate(filt_view):
                filt_view[i] = step((n - i) * dt, signal_in)

        return filtered

    def filtfilt(self, x, **kwargs):
        """Zero-phase filtering of ``x`` using this filter.

        Equivalent to `filt(x, filtfilt=True, **kwargs) <.Synapse.filt>`.
        """
        return self.filt(x, filtfilt=True, **kwargs)


class LinearFilter(Synapse):
    """General linear time-invariant (LTI) system synapse.

    This class can be used to implement any linear filter, given the
    filter's transfer function. [1]_

    Parameters
    ----------
    num : array_like
        Numerator coefficients of transfer function.
    den : array_like
        Denominator coefficients of transfer function.
    analog : boolean, optional
        Whether the synapse coefficients are analog (i.e. continuous-time),
        or discrete. Analog coefficients will be converted to discrete for
        simulation using the simulator ``dt``.
    method : string
        The method to use for discretization (if ``analog`` is True). See
        `scipy.signal.cont2discrete` for information about the options.

        .. versionadded:: 3.0.0

    Attributes
    ----------
    analog : boolean
        Whether the synapse coefficients are analog (i.e. continuous-time),
        or discrete. Analog coefficients will be converted to discrete for
        simulation using the simulator ``dt``.
    den : ndarray
        Denominator coefficients of transfer function.
    num : ndarray
        Numerator coefficients of transfer function.
    method : string
        The method to use for discretization (if ``analog`` is True). See
        `scipy.signal.cont2discrete` for information about the options.

    References
    ----------
    .. [1] https://en.wikipedia.org/wiki/Filter_%28signal_processing%29
    """

    num = NdarrayParam("num", shape="*")
    den = NdarrayParam("den", shape="*")
    analog = BoolParam("analog")
    method = EnumParam(
        "method", values=("gbt", "bilinear", "euler", "backward_diff", "zoh")
    )

    def __init__(self, num, den, analog=True, method="zoh", **kwargs):
        super().__init__(**kwargs)
        self.num = num
        self.den = den
        self.analog = analog
        self.method = method

    def combine(self, obj):
        """Combine in series with another LinearFilter."""
        if not isinstance(obj, LinearFilter):
            raise ValidationError(
                "Can only combine with other LinearFilters", attr="obj"
            )
        if self.analog != obj.analog:
            raise ValidationError(
                "Cannot combine analog and digital filters", attr="obj"
            )
        num = np.polymul(self.num, obj.num)
        den = np.polymul(self.den, obj.den)
        return LinearFilter(
            num,
            den,
            analog=self.analog,
            default_size_in=self.default_size_in,
            default_size_out=self.default_size_out,
            default_dt=self.default_dt,
            seed=self.seed,
        )

    def evaluate(self, frequencies):
        """Evaluate the transfer function at the given frequencies.

        Examples
        --------
        Using the ``evaluate`` function to make a Bode plot:

        .. testcode::

           import matplotlib.pyplot as plt

           synapse = nengo.synapses.LinearFilter([1], [0.02, 1])
           f = np.logspace(-1, 3, 100)
           y = synapse.evaluate(f)
           plt.subplot(211); plt.semilogx(f, 20*np.log10(np.abs(y)))
           plt.xlabel('frequency [Hz]'); plt.ylabel('magnitude [dB]')
           plt.subplot(212); plt.semilogx(f, np.angle(y))
           plt.xlabel('frequency [Hz]'); plt.ylabel('phase [radians]')
        """
        frequencies = 2.0j * np.pi * frequencies
        w = frequencies if self.analog else np.exp(frequencies)
        y = np.polyval(self.num, w) / np.polyval(self.den, w)
        return y

    def _get_ss(self, dt):
        A, B, C, D = tf2ss(self.num, self.den)

        # discretize (if len(A) == 0, filter is stateless and already discrete)
        if self.analog and len(A) > 0:
            A, B, C, D, _ = cont2discrete((A, B, C, D), dt, method=self.method)

        return A, B, C, D

    def make_state(self, shape_in, shape_out, dt, dtype=None, y0=0):
        assert shape_in == shape_out

        dtype = rc.float_dtype if dtype is None else np.dtype(dtype)
        if dtype.kind != "f":
            raise ValidationError(
                f"Only float data types are supported (got {dtype}). Please cast "
                "your data to a float type.",
                attr="dtype",
                obj=self,
            )

        A, B, C, D = self._get_ss(dt)

        # create state memory variable X
        X = np.zeros((A.shape[0],) + shape_out, dtype=dtype)

        # initialize X using y0 as steady-state output
        y0 = np.array(y0, copy=False, ndmin=2)
        if (y0 == 0).all():
            # just leave X as zeros in this case, so that this value works
            # for unstable systems
            pass
        elif LinearFilter.OneX.check(A, B, C, D, X):
            # OneX combines B and C into one scaling value `b`
            b = B.item() * C.item()
            X[:] = (b / (1 - A.item())) * y0
        else:
            # Solve for u0 (input) given y0 (output), then X given u0
            assert B.ndim == 1 or B.ndim == 2 and B.shape[1] == 1
            y0 = np.array(y0, copy=False, ndmin=2)
            IAB = np.linalg.solve(np.eye(len(A)) - A, B)
            Q = C.dot(IAB) + D  # multiplier from input to output (DC gain)
            assert Q.size == 1
            if np.abs(Q.item()) > 1e-8:
                u0 = y0 / Q.item()
                X[:] = IAB.dot(u0)
            else:
                raise ValidationError(
                    "Cannot solve for state if DC gain is zero. Please set `y0=0`.",
                    "y0",
                    obj=self,
                )

        return {"X": X}

    def make_step(self, shape_in, shape_out, dt, rng, state):
        """Returns a `.Step` instance that implements the linear filter."""
        assert shape_in == shape_out
        assert state is not None

        A, B, C, D = self._get_ss(dt)
        X = state["X"]

        if LinearFilter.NoX.check(A, B, C, D, X):
            return LinearFilter.NoX(A, B, C, D, X)
        if LinearFilter.OneXScalar.check(A, B, C, D, X):
            return LinearFilter.OneXScalar(A, B, C, D, X)
        elif LinearFilter.OneX.check(A, B, C, D, X):
            return LinearFilter.OneX(A, B, C, D, X)
        elif LinearFilter.NoD.check(A, B, C, D, X):
            return LinearFilter.NoD(A, B, C, D, X)
        else:
            assert LinearFilter.General.check(A, B, C, D, X)
            return LinearFilter.General(A, B, C, D, X)

    class Step:
        """Abstract base class for LTI filtering step functions."""

        def __init__(self, A, B, C, D, X):
            if not self.check(A, B, C, D, X):
                raise ValidationError(
                    "Matrices do not meet the requirements for this Step",
                    attr="A,B,C,D,X",
                    obj=self,
                )
            self.A = A
            self.B = B
            self.C = C
            self.D = D
            self.X = X

        def __call__(self, t, signal):
            raise NotImplementedError("Step object must implement __call__")

        @classmethod
        def check(cls, A, B, C, D, X):
            if A.size == 0:
                return X.size == B.size == C.size == 0 and D.size == 1
            else:
                return (
                    A.shape[0] == A.shape[1] == B.shape[0] == C.shape[1]
                    and A.shape[0] == X.shape[0]
                    and C.shape[0] == B.shape[1] == 1
                    and D.size == 1
                )

    class NoX(Step):
        """Step for system with no state, only passthrough matrix (D)."""

        def __init__(self, A, B, C, D, X):
            super().__init__(A, B, C, D, X)
            self.d = D.item()

        def __call__(self, t, signal):
            return self.d * signal

        @classmethod
        def check(cls, A, B, C, D, X):
            return super().check(A, B, C, D, X) and A.size == 0

    class OneX(Step):
        """Step for systems with one state element and no passthrough (D)."""

        def __init__(self, A, B, C, D, X):
            super().__init__(A, B, C, D, X)
            self.a = A.item()
            self.b = C.item() * B.item()

        def __call__(self, t, signal):
            self.X *= self.a
            self.X += self.b * signal
            return self.X[0]

        @classmethod
        def check(cls, A, B, C, D, X):
            return super().check(A, B, C, D, X) and (len(A) == 1 and (D == 0).all())

    class OneXScalar(OneX):
        """Step for systems with one state element, no passthrough, and a size-1 input.

        Using the builtin float math improves performance.
        """

        def __call__(self, t, signal):
            self.X[:] = self.a * self.X.item() + self.b * signal.item()
            return self.X[0]

        @classmethod
        def check(cls, A, B, C, D, X):
            return super().check(A, B, C, D, X) and X.size == 1

    class NoD(Step):
        """Step for systems with no passthrough matrix (D).

        Implements::

            x[t] = A x[t-1] + B u[t]
            y[t] = C x[t]

        Note how the input has been advanced one step as compared with the
        General system below, to remove the unnecessary delay.
        """

        def __call__(self, t, signal):
            self.X[:] = np.dot(self.A, self.X) + self.B * signal
            return np.dot(self.C, self.X)[0]

        @classmethod
        def check(cls, A, B, C, D, X):
            return super().check(A, B, C, D, X) and (len(A) >= 1 and (D == 0).all())

    class General(Step):
        """Step for any LTI system with at least one state element (X).

        Implements::

            x[t+1] = A x[t] + B u[t]
            y[t] = C x[t] + D u[t]

        Use ``NoX`` for systems with no state elements.
        """

        def __call__(self, t, signal):
            Y = np.dot(self.C, self.X)[0] + self.D * signal
            self.X[:] = np.dot(self.A, self.X) + self.B * signal
            return Y

        @classmethod
        def check(cls, A, B, C, D, X):
            return super().check(A, B, C, D, X) and len(A) >= 1


class Lowpass(LinearFilter):
    """Standard first-order lowpass filter synapse.

    The impulse-response function is given by::

        f(t) = (1 / tau) * exp(-t / tau)

    Parameters
    ----------
    tau : float
        The time constant of the filter in seconds.

    Attributes
    ----------
    tau : float
        The time constant of the filter in seconds.
    """

    tau = NumberParam("tau", low=0)

    def __init__(self, tau, **kwargs):
        super().__init__([1], [tau, 1], **kwargs)
        self.tau = tau


class Alpha(LinearFilter):
    """Alpha-function filter synapse.

    The impulse-response function is given by::

        alpha(t) = (t / tau**2) * exp(-t / tau)

    and was found by [1]_ to be a good basic model for synapses.

    Parameters
    ----------
    tau : float
        The time constant of the filter in seconds.

    Attributes
    ----------
    tau : float
        The time constant of the filter in seconds.

    References
    ----------
    .. [1] Mainen, Z.F. and Sejnowski, T.J. (1995). Reliability of spike timing
       in neocortical neurons. Science (New York, NY), 268(5216):1503-6.
    """

    tau = NumberParam("tau", low=0)

    def __init__(self, tau, **kwargs):
        super().__init__([1], [tau ** 2, 2 * tau, 1], **kwargs)
        self.tau = tau


class Triangle(Synapse):
    """Triangular finite impulse response (FIR) synapse.

    This synapse has a triangular and finite impulse response. The length of
    the triangle is ``t`` seconds; thus the digital filter will have
    ``t / dt + 1`` taps.

    Parameters
    ----------
    t : float
        Length of the triangle, in seconds.

    Attributes
    ----------
    t : float
        Length of the triangle, in seconds.
    """

    t = NumberParam("t", low=0)

    def __init__(self, t, **kwargs):
        super().__init__(**kwargs)
        self.t = t

    def _get_coefficients(self, dt, dtype=None):
        dtype = rc.float_dtype if dtype is None else np.dtype(dtype)

        n_taps = int(np.round(self.t / float(dt))) + 1
        num = np.arange(n_taps, 0, -1, dtype=rc.float_dtype)
        num /= num.sum()

        # Minimal multiply implementation finds the difference between
        # coefficients and subtracts a scaled signal at each time step.
        n0, ndiff = num[0], num[-1]

        return n_taps, n0, ndiff

    def make_state(self, shape_in, shape_out, dt, dtype=None, y0=0):
        assert shape_in == shape_out
        dtype = rc.float_dtype if dtype is None else np.dtype(dtype)

        n_taps, _, ndiff = self._get_coefficients(dt, dtype=dtype)
        Y = np.zeros(shape_out, dtype=dtype)
        X = np.zeros((n_taps,) + shape_out, dtype=dtype)
        Xi = np.zeros(1, dtype=dtype)  # counter for X position

        if y0 != 0 and len(X) > 0:
            y0 = np.array(y0, copy=False, ndmin=1)
            X[:] = ndiff * y0[None, ...]
            Y[:] = y0

        return {"Y": Y, "X": X, "Xi": Xi}

    def make_step(self, shape_in, shape_out, dt, rng, state):
        assert shape_in == shape_out
        assert state is not None

        Y, X, Xi = state["Y"], state["X"], state["Xi"]
        n_taps, n0, ndiff = self._get_coefficients(dt, dtype=Y.dtype)
        assert len(X) == n_taps

        def step_triangle(t, signal):
            Y[...] += n0 * signal
            Y[...] -= X.sum(axis=0)
            Xi[:] = (Xi + 1) % len(X)
            X[int(Xi.item())] = ndiff * signal
            return Y

        return step_triangle


class SynapseParam(Parameter):
    equatable = True

    def __init__(self, name, default=Unconfigurable, optional=True, readonly=None):
        super().__init__(name, default, optional, readonly)

    def coerce(self, instance, synapse):  # pylint: disable=arguments-renamed
        synapse = Lowpass(synapse) if is_number(synapse) else synapse
        self.check_type(instance, synapse, Synapse)
        return super().coerce(instance, synapse)



import numpy as np

import nengo
from nengo.dists import Uniform
from nengo.spa.action_objects import Convolution, Source, Symbol
from nengo.spa.module import Module


class Thalamus(nengo.networks.Thalamus, Module):
    """A thalamus, implementing the effects for an associated basal ganglia.

    See `.spa.BasalGanglia` for information on the basal ganglia, and
    `.networks.Thalamus` for details on the underlying network.

    Parameters
    ----------
    bg : spa.BasalGanglia
        The associated basal ganglia that defines the action to implement.
    neurons_action : int, optional
        Number of neurons per action to represent the selection.
    threshold_action : float, optional
        Minimum value for action representation.
    mutual_inhibit : float, optional
        Strength of inhibition between actions.
    route_inhibit : float, optional
        Strength of inhibition for unchosen actions.
    synapse_inhibit : float, optional
        Synaptic filter to apply for inhibition between actions.
    synapse_bg : float, optional
        Synaptic filter for connection between basal ganglia and thalamus.
    synapse_direct : float, optional
        Synaptic filter for direct outputs.
    neurons_channel_dim : int, optional
        Number of neurons per routing channel dimension.
    subdim_channel : int, optional
        Number of subdimensions used in routing channel.
    synapse_channel : float, optional
        Synaptic filter for channel inputs and outputs.
    neurons_cconv : int, optional
        Number of neurons per circular convolution dimension.
    neurons_gate : int, optional
        Number of neurons per gate.
    threshold_gate : float, optional
        Minimum value for gating neurons.
    synapse_to-gate : float, optional
        Synaptic filter for controlling a gate.

    label : str, optional
        A name for the ensemble. Used for debugging and visualization.
    seed : int, optional
        The seed used for random number generation.
    add_to_container : bool, optional
        Determines if this Network will be added to the current container.
        If None, will be true if currently within a Network.
    """

    def __init__(
        self,
        bg,
        neurons_action=50,
        threshold_action=0.2,
        mutual_inhibit=1.0,
        route_inhibit=3.0,
        synapse_inhibit=0.008,
        synapse_bg=0.008,
        synapse_direct=0.01,
        neurons_channel_dim=50,
        subdim_channel=16,
        synapse_channel=0.01,
        neurons_cconv=200,
        neurons_gate=40,
        threshold_gate=0.3,
        synapse_to_gate=0.002,
        label=None,
        seed=None,
        add_to_container=None,
    ):

        self.bg = bg
        self.neurons_action = neurons_action
        self.mutual_inhibit = mutual_inhibit
        self.route_inhibit = route_inhibit
        self.synapse_inhibit = synapse_inhibit
        self.synapse_direct = synapse_direct
        self.threshold_action = threshold_action
        self.neurons_channel_dim = neurons_channel_dim
        self.subdim_channel = subdim_channel
        self.synapse_channel = synapse_channel
        self.neurons_gate = neurons_gate
        self.neurons_cconv = neurons_cconv
        self.threshold_gate = threshold_gate
        self.synapse_to_gate = synapse_to_gate
        self.synapse_bg = synapse_bg
        self.spa = None

        self.gates = {}  # gating ensembles per action (created as needed)
        self.channels = {}  # channels to pass transformed data between modules

        Module.__init__(self)
        nengo.networks.Thalamus.__init__(
            self,
            dimensions=self.bg.actions.count,
            n_neurons_per_ensemble=self.neurons_action,
            mutual_inhib=self.mutual_inhibit,
            threshold=self.threshold_action,
            label=label,
            seed=seed,
            add_to_container=add_to_container,
        )

    def on_add(self, spa):
        Module.on_add(self, spa)
        self.spa = spa

        with spa:
            # connect basal ganglia to thalamus
            nengo.Connection(
                self.bg.output, self.actions.input, synapse=self.synapse_bg
            )

        # implement the various effects
        for i, action in enumerate(self.bg.actions.actions):
            for name, effects in action.effect.effect.items():
                for effect in effects.expression.items:
                    if isinstance(effect, (int, float)):
                        effect = Symbol("%g" % effect)
                    if isinstance(effect, Symbol):
                        self.add_direct_effect(i, name, effect.symbol)
                    elif isinstance(effect, Source):
                        self.add_route_effect(
                            i,
                            name,
                            effect.name,
                            effect.transform.symbol,
                            effect.inverted,
                        )
                    elif isinstance(effect, Convolution):
                        self.add_conv_effect(i, name, effect)
                    else:
                        raise NotImplementedError(
                            "Subexpression '%s' from action '%s' is not "
                            "supported by the Thalamus." % (effect, action)
                        )

    def add_direct_effect(self, index, target_name, value):
        """Cause an action to drive a particular module input to value.

        Parameters
        ----------
        index : int
            The action number that causes this effect.
        target_name : str
            The name of the module input to connect to.
        value : str
            A semantic pointer to be sent into the module when this action
            is active.
        """
        sink, vocab = self.spa.get_module_input(target_name)
        transform = np.array([vocab.parse(value).v]).T

        with self.spa:
            nengo.Connection(
                self.actions.ensembles[index],
                sink,
                transform=transform,
                synapse=self.synapse_direct,
            )

    def get_gate(self, index, target_name):
        """Return the gate for an action.

        The gate will be created if it does not already exist. The gate
        neurons have no activity when the action is selected, but are
        active when the action is not selected. This makes the gate useful
        for inhibiting ensembles that should only be active when this
        action is active.
        """

        target_module = self.spa.get_module(target_name)

        if index not in self.gates:
            with target_module:
                intercepts = Uniform(self.threshold_gate, 1)
                gate = nengo.Ensemble(
                    self.neurons_gate,
                    dimensions=1,
                    intercepts=intercepts,
                    label="gate[%d]" % index,
                    encoders=[[1]] * self.neurons_gate,
                )
                if not hasattr(target_module, "bias"):
                    target_module.bias = nengo.Node([1], label=target_name + " bias")
                nengo.Connection(target_module.bias, gate, synapse=None)

            with self.spa:
                nengo.Connection(
                    self.actions.ensembles[index],
                    gate,
                    synapse=self.synapse_to_gate,
                    transform=-1,
                )

            with self:
                self.gates[index] = gate
        return self.gates[index]

    def add_route_effect(self, index, target_name, source_name, transform, inverted):
        """Set an action to send source to target with the given transform.

        Parameters
        ----------
        index : int
            The action number that will cause this effect.
        target_name : str
            The name of the module input to affect.
        source_name : str
            The name of the module output to read from. If this output uses
            a different Vocabulary than the target, a linear transform
            will be applied to convert from one to the other.
        transform : str
            A semantic point to convolve with the source value before
            sending it into the target. This transform takes
            place in the source Vocabulary.
        inverted : bool
            Whether to perform inverse convolution on the source.
        """
        with self:
            gate = self.get_gate(index, target_name)

            target, target_vocab = self.spa.get_module_input(target_name)
            source, source_vocab = self.spa.get_module_output(source_name)

            target_module = self.spa.get_module(target_name)
            source_module = self.spa.get_module(source_name)

            # build a communication channel between the source and target
            dim = target_vocab.dimensions

            # Determine size of subdimension. If target module is spa.Buffer,
            # use target module's subdimension. Otherwise use default
            # (self.subdim_channel).
            # TODO: Use the ensemble properties of target module instead?
            #       - How to get these properties?
            subdim = self.subdim_channel
            if isinstance(target_module, nengo.spa.Buffer):
                subdim = target_module.state.dimensions_per_ensemble
            elif isinstance(source_module, nengo.spa.Buffer):
                subdim = source_module.state.dimensions_per_ensemble
            elif dim < subdim:
                subdim = dim
            elif dim % subdim != 0:
                subdim = 1

        with target_module:
            channel = nengo.networks.EnsembleArray(
                self.neurons_channel_dim * subdim,
                dim // subdim,
                ens_dimensions=subdim,
                radius=np.sqrt(float(subdim) / dim),
                label="channel_%d_%s" % (index, target_name),
            )

        with self.spa:
            # inhibit the channel when the action is not chosen
            inhibit = [[-self.route_inhibit]] * (self.neurons_channel_dim * subdim)
            for e in channel.ensembles:
                nengo.Connection(
                    gate, e.neurons, transform=inhibit, synapse=self.synapse_inhibit
                )

            # compute the requested transform
            t = source_vocab.parse(transform).get_convolution_matrix()
            if inverted:
                D = source_vocab.dimensions
                t = np.dot(t, np.eye(D)[-np.arange(D)])
            # handle conversion between different Vocabularies
            if target_vocab is not source_vocab:
                t = np.dot(source_vocab.transform_to(target_vocab), t)

            # connect source to target
            nengo.Connection(
                source, channel.input, transform=t, synapse=self.synapse_channel
            )
            nengo.Connection(channel.output, target, synapse=self.synapse_channel)

    def add_conv_effect(self, index, target_name, effect):
        """Set an action to combine two sources and send to target.

        Parameters
        ----------
        index : int
            The action number that will cause this effect.
        target_name : str
            The name of the module input to affect.
        effect : Convolution
            The details of the convolution to implement.
        """
        target_module = self.spa.get_module(target_name)
        cconv = nengo.spa.action_build.convolution(
            self, target_name, effect, self.neurons_cconv, self.synapse_channel
        )

        gate = self.get_gate(index, target_name)

        with target_module:
            # inhibit the convolution when the action is not chosen
            for e in cconv.product.all_ensembles:
                inhibit = -np.ones((e.n_neurons, 1)) * self.route_inhibit
                nengo.Connection(
                    gate, e.neurons, transform=inhibit, synapse=self.synapse_inhibit
                )



import pytest

import nengo
from nengo.exceptions import BuildError


def test_build_node_error(Simulator):
    """Tests a build error for the build_node function"""

    class BadOutputType:
        pass

    with nengo.Network() as net:
        n = nengo.Node(0)

    # hack to change node type without tripping API validation
    nengo.Node.output.data[n] = BadOutputType()

    with pytest.raises(BuildError, match="Invalid node output type"):
        with Simulator(net):
            pass



from .magic import decorator


@decorator
def with_self(method, network, args, kwargs):
    """Wraps a method with ``with network:``.

    This makes it easy to add methods to a network that create new
    Nengo objects. Instead of writing ``with self`` at the top of the method
    and indenting everything over, you can instead use this decorator.

    Examples
    --------
    The two methods in the following class do the same thing:

    .. testcode::

       from nengo.utils.network import with_self

       class MyNetwork(nengo.Network):
           def add_one_1(self):
               with self:
                   node = nengo.Node(output=1)

           @with_self
           def add_one_2(self):
               node = nengo.Node(output=1)
    """
    with network:
        return method(*args, **kwargs)


def activate_direct_mode(network):
    """Activates direct mode for a network.

    This sets the neuron type of all ensembles to a `nengo.Direct`
    instance unless:

    - there is a connection to or from the ensemble's neurons
    - there is a probe on an ensemble's neurons
    - the ensemble has a connection with a learning rule attached.

    Parameters
    ----------
    network : Network
        Network to activate direct mode for.
    """
    # pylint: disable=import-outside-toplevel
    # imported here to avoid circular imports
    from nengo.ensemble import Neurons
    from nengo.neurons import Direct

    requires_neurons = set()

    for c in network.all_connections:
        if isinstance(c.pre_obj, Neurons):
            requires_neurons.add(c.pre_obj.ensemble)
        if isinstance(c.post_obj, Neurons):
            requires_neurons.add(c.post_obj.ensemble)
        if c.learning_rule_type is not None:
            requires_neurons.add(c.pre_obj)
            requires_neurons.add(c.post_obj)

    for p in network.all_probes:
        if isinstance(p.obj, Neurons):
            requires_neurons.add(p.obj.ensemble)

    for e in network.all_ensembles:
        if e not in requires_neurons:
            e.neuron_type = Direct()



import importlib
import itertools

import numpy as np
import pytest

import nengo.utils.numpy as npext
from nengo._vendor.scipy import expm
from nengo.exceptions import ValidationError
from nengo.utils.numpy import array, array_hash, as_shape, broadcast_shape, meshgrid_nd


def test_meshgrid_nd(allclose):
    a = [0, 0, 1]
    b = [1, 2, 3]
    c = [23, 42]
    expected = [
        np.array(
            [
                [[0, 0], [0, 0], [0, 0]],
                [[0, 0], [0, 0], [0, 0]],
                [[1, 1], [1, 1], [1, 1]],
            ]
        ),
        np.array(
            [
                [[1, 1], [2, 2], [3, 3]],
                [[1, 1], [2, 2], [3, 3]],
                [[1, 1], [2, 2], [3, 3]],
            ]
        ),
        np.array(
            [
                [[23, 42], [23, 42], [23, 42]],
                [[23, 42], [23, 42], [23, 42]],
                [[23, 42], [23, 42], [23, 42]],
            ]
        ),
    ]
    actual = meshgrid_nd(a, b, c)
    assert allclose(expected, actual)


@pytest.mark.parametrize("nnz", [7, 300])
def test_array_hash_sparse(nnz, rng):
    scipy_sparse = pytest.importorskip("scipy.sparse")

    if nnz == 7:
        shape = (5, 5)
        idxs_a = ([0, 0, 1, 2, 3, 3, 4], [0, 2, 3, 4, 2, 4, 0])
        idxs_b = ([0, 1, 1, 2, 3, 3, 4], [1, 2, 3, 4, 2, 4, 0])

        data_a = [1.0, 2.0, 1.5, 2.3, 1.2, 2.5, 1.8]
        data_b = [1.0, 1.0, 1.5, 2.3, 1.2, 2.5, 1.8]
    else:
        shape = (100, 100)

        idxs_a = np.unravel_index(rng.permutation(np.prod(shape))[:nnz], shape)
        idxs_b = np.unravel_index(rng.permutation(np.prod(shape))[:nnz], shape)

        data_a = rng.uniform(-1, 1, size=nnz)
        data_b = rng.uniform(-1, 1, size=nnz)

    matrices = [[] for _ in range(6)]

    for (rows, cols), data in itertools.product((idxs_a, idxs_b), (data_a, data_b)):

        csr = scipy_sparse.csr_matrix((data, (rows, cols)), shape=shape)
        matrices[0].append(csr)
        matrices[1].append(csr.tocsc())
        matrices[2].append(csr.tocoo())
        matrices[3].append(csr.tobsr())
        matrices[4].append(csr.todok())
        matrices[5].append(csr.tolil())
        # matrices[6].append(csr.todia())  # warns about inefficiency

    # ensure hash is reproducible
    for matrix in (m for kind in matrices for m in kind):
        assert array_hash(matrix) == array_hash(matrix)

    # ensure hash is different for different matrices
    for kind in matrices:
        hashes = [array_hash(matrix) for matrix in kind]
        assert len(np.unique(hashes)) == len(
            kind
        ), f"Different matrices should have different hashes: {hashes}"


def test_expm(rng, allclose):
    scipy_linalg = pytest.importorskip("scipy.linalg")
    for a in [np.eye(3), rng.randn(10, 10), -10 + rng.randn(10, 10)]:
        assert allclose(scipy_linalg.expm(a), expm(a))


def test_as_shape_errors():
    """Tests errors generated by the `as_shape` function"""
    with pytest.raises(ValueError, match="cannot be safely converted to a shape"):
        as_shape(1.0)  # float is noniterable and noninteger


def test_brodcast_shape():
    assert broadcast_shape(shape=(3, 2), length=3) == (1, 3, 2)
    assert broadcast_shape(shape=(3, 2), length=4) == (1, 1, 3, 2)
    assert broadcast_shape(shape=(3, 2), length=2) == (3, 2)


def test_array():
    assert array([1, 2, 3], dims=4).shape == (3, 1, 1, 1)
    assert array([1, 2, 3], dims=1).shape == (3,)
    assert array([1, 2, 3], min_dims=2).shape == (3, 1)
    assert array([1, 2, 3], min_dims=1).shape == (3,)

    x = array([1, 2, 3], readonly=True)
    with pytest.raises(ValueError, match="read-only"):
        x[0] = 3

    with pytest.raises(ValidationError, match="Input cannot be cast to array"):
        array([[1, 2, 3]], dims=1)


def test_rfftfreq_fallback(monkeypatch):
    np_rfftfreq = np.fft.rfftfreq
    monkeypatch.delattr(np.fft, "rfftfreq")
    importlib.reload(npext)

    for n in (3, 4, 8, 9):
        for d in (1.0, 3.4, 9.8):
            assert np.allclose(npext.rfftfreq(n, d=d), np_rfftfreq(n, d=d))



import nengo
from nengo import spa


def test_basic():
    with spa.SPA() as model:
        model.compare = spa.Compare(dimensions=16)

    inputA = model.get_module_input("compare_A")
    inputB = model.get_module_input("compare_B")
    output = model.get_module_output("compare")
    # all nodes should be acquired correctly
    assert inputA[0] is model.compare.inputA
    assert inputB[0] is model.compare.inputB
    assert output[0] is model.compare.output
    # all inputs should share the same vocab
    assert inputA[1] is inputB[1]
    assert inputA[1].dimensions == 16
    # output should have no vocab
    assert output[1] is None


def test_run(Simulator, seed):
    with spa.SPA(seed=seed) as model:
        model.compare = spa.Compare(dimensions=16)

        def inputA(t):
            if 0 <= t < 0.1:
                return "A"
            else:
                return "B"

        model.input = spa.Input(compare_A=inputA, compare_B="A")

    compare, vocab = model.get_module_output("compare")

    with model:
        p = nengo.Probe(compare, "output", synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.2)

    assert sim.data[p][100] > 0.8
    assert sim.data[p][199] < 0.2



import numpy as np

import nengo
import nengo.spa.action_build
from nengo.spa.action_objects import Convolution, Source, Symbol
from nengo.spa.module import Module


class Cortical(Module):
    """A SPA module for forming connections between other modules.

    Parameters
    ----------
    actions : Actions
        The actions to implement.
    synapse : float, optional
        The synaptic filter to use for the connections.
    neurons_cconv : int, optional
        Number of neurons per circular convolution dimension.

    label : str, optional
        A name for the ensemble. Used for debugging and visualization.
    seed : int, optional
        The seed used for random number generation.
    add_to_container : bool, optional
        Determines if this Network will be added to the current container.
        If None, will be true if currently within a Network.
    """

    def __init__(
        self,
        actions,
        synapse=0.01,
        neurons_cconv=200,
        label=None,
        seed=None,
        add_to_container=None,
    ):
        super().__init__(label, seed, add_to_container)
        self.actions = actions
        self.synapse = synapse
        self.neurons_cconv = neurons_cconv
        self.spa = None

    def on_add(self, spa):
        Module.on_add(self, spa)
        self.spa = spa

        # parse the provided class and match it up with the spa model
        self.actions.process(spa)
        for action in self.actions.actions:
            if action.condition is not None:
                raise NotImplementedError(
                    "Cortical actions do not support "
                    "conditional expressions: %s." % action.condition
                )
            for name, effects in action.effect.effect.items():
                for effect in effects.expression.items:
                    if isinstance(effect, Symbol):
                        self.add_direct_effect(name, effect.symbol)
                    elif isinstance(effect, Source):
                        self.add_route_effect(
                            name, effect.name, effect.transform.symbol, effect.inverted
                        )
                    elif isinstance(effect, Convolution):
                        self.add_conv_effect(name, effect)
                    else:
                        raise NotImplementedError(
                            "Subexpression '%s' from action '%s' is not "
                            "supported by the cortex." % (effect, action)
                        )

    def add_direct_effect(self, target_name, value):
        """Make a fixed constant input to a module.

        Parameters
        ----------
        target_name : str
            The name of the module input to use.
        value : str
            A semantic pointer to be sent to the module input.
        """
        target_module = self.spa.get_module(target_name)
        sink, vocab = self.spa.get_module_input(target_name)
        transform = np.array([vocab.parse(value).v]).T

        with target_module:
            if not hasattr(target_module, "bias"):
                target_module.bias = nengo.Node([1], label=target_name + " bias")
            nengo.Connection(
                target_module.bias, sink, transform=transform, synapse=self.synapse
            )

    def add_route_effect(self, target_name, source_name, transform, inverted):
        """Connect a module output to a module input.

        Parameters
        ----------
        target_name : str
            The name of the module input to effect.
        source_name : str
            The name of the module output to read from. If this output uses
            a different vocabulary than the target, a linear transform
            will be applied to convert from one to the other.
        transform : str
            A semantic pointer to convolve with the source value before
            sending it into the target. This transform takes
            place in the source vocabulary.
        inverted : bool
            Whether to invert the transform.
        """
        target, target_vocab = self.spa.get_module_input(target_name)
        source, source_vocab = self.spa.get_module_output(source_name)

        t = source_vocab.parse(transform).get_convolution_matrix()
        if inverted:
            D = source_vocab.dimensions
            t = np.dot(t, np.eye(D)[-np.arange(D)])

        if target_vocab is not source_vocab:
            t = np.dot(source_vocab.transform_to(target_vocab), t)

        with self.spa:
            nengo.Connection(source, target, transform=t, synapse=self.synapse)

    def add_conv_effect(self, target_name, effect):
        """Convolve the output of two modules and send result to target.

        Parameters
        ----------
        target_name : str
            The name of the module input to affect
        effect : Convolution
            The details of the convolution to implement.
        """
        nengo.spa.action_build.convolution(
            self, target_name, effect, self.neurons_cconv, self.synapse
        )



import os
import subprocess
import sys
import textwrap

import pytest

import nengo
from nengo.builder.builder import Builder


def test_seed_override(seed, allclose):
    """Test that seeds are not overwritten by the seeding function"""
    with nengo.Network(seed=seed - 1) as net:
        a = nengo.Ensemble(10, 1, seed=seed - 2)
        b = nengo.Ensemble(10, 1, seed=seed + 2)

    model = nengo.builder.Model()
    model.seeds[net] = seed + 1
    model.seeds[a] = seed + 2

    # note: intentionally setting this to the 'wrong' value, to check that
    # it isn't being overridden (things with seeds set should have seeded=True)
    model.seeded[net] = False
    model.seeded[a] = False

    model.build(net)

    assert model.seeds[net] == seed + 1
    assert model.seeds[a] == seed + 2
    assert not model.seeded[net]
    assert not model.seeded[a]
    assert allclose(model.params[a].gain, model.params[b].gain)


def test_build_twice():
    """Tests warning for building an object twice"""
    model = nengo.builder.Model()
    ens = nengo.Ensemble(10, 1, add_to_container=False)
    model.seeds[ens] = 0
    model.build(ens)
    built_ens = model.params[ens]

    with pytest.warns(UserWarning, match="has already been built"):
        assert model.build(ens) is None
    assert model.params[ens] is built_ens


def test_register_builder_twice_warning():
    """Tests warning for registering a builder twice"""

    class Test:
        pass

    my_builder = Builder.register(Test)
    my_builder(1)
    with pytest.warns(Warning, match="Type .* already has a builder. Overwriting"):
        my_builder(1)  # repeat setup warning


@pytest.mark.skipif(sys.version_info < (3, 6, 0), reason="requires ordered dicts")
def test_deterministic_op_order(tmp_path):
    code = textwrap.dedent(
        """
    import nengo

    with nengo.Network(seed=0) as net:
        # use ensemblearrays as they create a lot of parallel ops
        ens0 = nengo.networks.EnsembleArray(1, 100)
        ens1 = nengo.networks.EnsembleArray(1, 100)
        nengo.Connection(ens0.output, ens1.input)
        nengo.Probe(ens1.output)

    # the optimizer uses WeakSets, which seem incompatible with ordering
    with nengo.Simulator(net, progress_bar=False, optimize=False) as sim:
        ops = sim.step_order

    for op in ops:
        print(type(op))
        for s in op.all_signals:
            print(s._name)
            print(s.shape)
            print(s.initial_value)
    """
    )
    tmp_path = tmp_path / "test.py"
    tmp_path.write_text(code, encoding="utf-8")

    env = os.environ.copy()

    env["PYTHONHASHSEED"] = "0"
    output0 = subprocess.run(
        [sys.executable, str(tmp_path)],
        stdout=subprocess.PIPE,
        env=env,
        encoding="utf-8",
        check=True,
    )

    env["PYTHONHASHSEED"] = "1"
    output1 = subprocess.run(
        [sys.executable, str(tmp_path)],
        stdout=subprocess.PIPE,
        env=env,
        encoding="utf-8",
        check=True,
    )

    assert output0.stdout == output1.stdout



import numpy as np

import nengo
from nengo.spa.module import Module


class Compare(Module):
    """A module for computing the dot product of two inputs.

    Parameters
    ----------
    dimensions : int
        Number of dimensions for the two vectors to be compared.
    vocab : Vocabulary, optional
        The vocabulary to use to interpret the vector. If None,
        the default vocabulary for the given dimensionality is used.
    neurons_per_multiply : int, optional
        Number of neurons to use in each product computation.
    input_magnitude : float, optional
        The expected magnitude of the vectors to be multiplied.
        This value is used to determine the radius of the ensembles
        computing the element-wise product.

    label : str, optional
        A name for the ensemble. Used for debugging and visualization.
    seed : int, optional
        The seed used for random number generation.
    add_to_container : bool, optional
        Determines if this Network will be added to the current container.
        If None, will be true if currently within a Network.
    """

    def __init__(
        self,
        dimensions,
        vocab=None,
        neurons_per_multiply=200,
        input_magnitude=1.0,
        label=None,
        seed=None,
        add_to_container=None,
    ):
        super().__init__(label, seed, add_to_container)
        if vocab is None:
            # use the default vocab for this number of dimensions
            vocab = dimensions

        with self:
            self.product = nengo.networks.Product(
                neurons_per_multiply, dimensions, input_magnitude=input_magnitude
            )

            self.inputA = nengo.Node(size_in=dimensions, label="inputA")
            self.inputB = nengo.Node(size_in=dimensions, label="inputB")
            self.output = nengo.Node(size_in=1, label="output")

        self.inputs = dict(A=(self.inputA, vocab), B=(self.inputB, vocab))
        self.outputs = dict(default=(self.output, None))

        with self:
            nengo.Connection(self.inputA, self.product.input_a, synapse=None)
            nengo.Connection(self.inputB, self.product.input_b, synapse=None)
            nengo.Connection(
                self.product.output, self.output, transform=np.ones((1, dimensions))
            )



"""Functions for filter design.

----

These are borrowed from SciPy and used under their license:

Copyright (c) 2001, 2002 Enthought, Inc.
All rights reserved.

Copyright (c) 2003-2012 SciPy Developers.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

  a. Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
  b. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  c. Neither the name of Enthought nor the names of the SciPy Developers
     may be used to endorse or promote products derived from this software
     without specific prior written permission.


THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
THE POSSIBILITY OF SUCH DAMAGE.
"""

import warnings

import numpy as np
from numpy import (
    allclose,
    array,
    asarray,
    atleast_1d,
    atleast_2d,
    dot,
    eye,
    poly,
    product,
    r_,
    roots,
    zeros,
)

from nengo._vendor.scipy import expm


class BadCoefficients(UserWarning):
    pass


def tf2zpk(b, a):
    """Return zero, pole, gain (z,p,k) representation from a numerator,
    denominator representation of a linear filter.

    Parameters
    ----------
    b : ndarray
        Numerator polynomial.
    a : ndarray
        Denominator polynomial.

    Returns
    -------
    z : ndarray
        Zeros of the transfer function.
    p : ndarray
        Poles of the transfer function.
    k : float
        System gain.

    Notes
    -----
    If some values of ``b`` are too close to 0, they are removed. In that case,
    a BadCoefficients warning is emitted.

    """
    b, a = normalize(b, a)
    b = (b + 0.0) / a[0]
    a = (a + 0.0) / a[0]
    k = b[0]
    b /= b[0]
    z = roots(b)
    p = roots(a)
    return z, p, k


def zpk2tf(z, p, k):
    """Return polynomial transfer function representation from zeros
    and poles

    Parameters
    ----------
    z : ndarray
        Zeros of the transfer function.
    p : ndarray
        Poles of the transfer function.
    k : float
        System gain.

    Returns
    -------
    b : ndarray
        Numerator polynomial.
    a : ndarray
        Denominator polynomial.

    """
    z = atleast_1d(z)
    k = atleast_1d(k)
    if len(z.shape) > 1:
        temp = poly(z[0])
        b = zeros((z.shape[0], z.shape[1] + 1), temp.dtype.char)
        if len(k) == 1:
            k = [k[0]] * z.shape[0]
        for i in range(z.shape[0]):
            b[i] = k[i] * poly(z[i])
    else:
        b = k * poly(z)
    a = atleast_1d(poly(p))
    return b, a


def normalize(b, a):
    """Normalize polynomial representation of a transfer function.

    If values of ``b`` are too close to 0, they are removed. In that case, a
    BadCoefficients warning is emitted.

    """
    b, a = map(atleast_1d, (b, a))
    if len(a.shape) != 1:
        raise ValueError("Denominator polynomial must be rank-1 array.")
    if len(b.shape) > 2:
        raise ValueError("Numerator polynomial must be rank-1 or rank-2 array.")
    if len(b.shape) == 1:
        b = asarray([b], b.dtype.char)
    while a[0] == 0.0 and len(a) > 1:
        a = a[1:]
    outb = b * (1.0) / a[0]
    outa = a * (1.0) / a[0]
    if allclose(0, outb[:, 0], atol=1e-14):
        warnings.warn(
            "Badly conditioned filter coefficients (numerator): the "
            "results may be meaningless",
            BadCoefficients,
        )
        while allclose(0, outb[:, 0], atol=1e-14) and (outb.shape[-1] > 1):
            outb = outb[:, 1:]
    if outb.shape[0] == 1:
        outb = outb[0]
    return outb, outa


def tf2ss(num, den):
    """Transfer function to state-space representation.

    Parameters
    ----------
    num, den : array_like
        Sequences representing the numerator and denominator polynomials.
        The denominator needs to be at least as long as the numerator.

    Returns
    -------
    A, B, C, D : ndarray
        State space representation of the system.

    """
    # Controller canonical state-space representation.
    #  if M+1 = len(num) and K+1 = len(den) then we must have M <= K
    #  states are found by asserting that X(s) = U(s) / D(s)
    #  then Y(s) = N(s) * X(s)
    #
    #   A, B, C, and D follow quite naturally.
    #
    num, den = normalize(num, den)  # Strips zeros, checks arrays
    nn = len(num.shape)
    if nn == 1:
        num = asarray([num], num.dtype)
    M = num.shape[1]
    K = len(den)
    if M > K:
        msg = "Improper transfer function. `num` is longer than `den`."
        raise ValueError(msg)
    if M == 0 or K == 0:  # Null system
        return array([], float), array([], float), array([], float), array([], float)

    # pad numerator to have same number of columns has denominator
    num = r_["-1", zeros((num.shape[0], K - M), num.dtype), num]

    if num.shape[-1] > 0:
        D = num[:, 0]
    else:
        D = array([], float)

    if K == 1:
        return array([], float), array([], float), array([], float), D

    frow = -1 * array([den[1:]])
    A = r_[frow, eye(K - 2, K - 1)]
    B = eye(K - 1, 1)
    C = num[:, 1:] - num[:, 0] * den[1:]
    return A, B, C, D


def _none_to_empty_2d(arg):
    if arg is None:
        return zeros((0, 0))
    else:
        return arg


def _atleast_2d_or_none(arg):
    if arg is not None:
        return atleast_2d(arg)
    return None


def _shape_or_none(M):
    if M is not None:
        return M.shape
    else:
        return (None,) * 2


def _choice_not_none(*args):
    for arg in args:
        if arg is not None:
            return arg
    return None


def _restore(M, shape):
    if M.shape == (0, 0):
        return zeros(shape)
    else:
        if M.shape != shape:
            raise ValueError("The input arrays have incompatible shapes.")
        return M


def abcd_normalize(A=None, B=None, C=None, D=None):
    """Check state-space matrices and ensure they are rank-2.

    If enough information on the system is provided, that is, enough
    properly-shaped arrays are passed to the function, the missing ones
    are built from this information, ensuring the correct number of
    rows and columns. Otherwise a ValueError is raised.

    Parameters
    ----------
    A, B, C, D : array_like, optional
        State-space matrices. All of them are None (missing) by default.

    Returns
    -------
    A, B, C, D : array
        Properly shaped state-space matrices.

    Raises
    ------
    ValueError
        If not enough information on the system was provided.

    """
    A, B, C, D = map(_atleast_2d_or_none, (A, B, C, D))

    Am, _ = _shape_or_none(A)
    Bm, Bn = _shape_or_none(B)
    Cm, Cn = _shape_or_none(C)
    Dm, Dn = _shape_or_none(D)

    p = _choice_not_none(Am, Bm, Cn)
    q = _choice_not_none(Bn, Dn)
    r = _choice_not_none(Cm, Dm)
    if p is None or q is None or r is None:
        raise ValueError("Not enough information on the system.")

    A, B, C, D = map(_none_to_empty_2d, (A, B, C, D))
    A = _restore(A, (p, p))
    B = _restore(B, (p, q))
    C = _restore(C, (r, p))
    D = _restore(D, (r, q))

    return A, B, C, D


def ss2tf(A, B, C, D, input=0):
    """State-space to transfer function.

    Parameters
    ----------
    A, B, C, D : ndarray
        State-space representation of linear system.
    input : int, optional
        For multiple-input systems, the input to use.

    Returns
    -------
    num : 2-D ndarray
        Numerator(s) of the resulting transfer function(s). ``num`` has one row
        for each of the system's outputs. Each row is a sequence representation
        of the numerator polynomial.
    den : 1-D ndarray
        Denominator of the resulting transfer function(s). ``den`` is a sequence
        representation of the denominator polynomial.

    """
    # transfer function is C (sI - A)**(-1) B + D
    A, B, C, D = map(asarray, (A, B, C, D))
    # Check consistency and make them all rank-2 arrays
    A, B, C, D = abcd_normalize(A, B, C, D)

    n_out, n_in = D.shape
    if input >= n_in:
        raise ValueError("System does not have the input specified.")

    # make MOSI from possibly MOMI system.
    if B.shape[-1] != 0:
        B = B[:, input]
    B.shape = (B.shape[0], 1)
    if D.shape[-1] != 0:
        D = D[:, input]

    try:
        den = poly(A)
    except ValueError:
        den = 1

    if (product(B.shape, axis=0) == 0) and (product(C.shape, axis=0) == 0):
        num = np.ravel(D)
        if (product(D.shape, axis=0) == 0) and (product(A.shape, axis=0) == 0):
            den = []
        return num, den

    num_states = A.shape[0]
    type_test = A[:, 0] + B[:, 0] + C[0, :] + D
    num = np.zeros((n_out, num_states + 1), type_test.dtype)
    for k in range(n_out):
        Ck = atleast_2d(C[k, :])
        num[k] = poly(A - dot(B, Ck)) + (D[k] - 1) * den

    return num, den


def zpk2ss(z, p, k):
    """Zero-pole-gain representation to state-space representation

    Parameters
    ----------
    z, p : sequence
        Zeros and poles.
    k : float
        System gain.

    Returns
    -------
    A, B, C, D : ndarray
        State-space matrices.

    """
    return tf2ss(*zpk2tf(z, p, k))


def ss2zpk(A, B, C, D, input=0):
    """State-space representation to zero-pole-gain representation.

    Parameters
    ----------
    A, B, C, D : ndarray
        State-space representation of linear system.
    input : int, optional
        For multiple-input systems, the input to use.

    Returns
    -------
    z, p : sequence
        Zeros and poles.
    k : float
        System gain.

    """
    return tf2zpk(*ss2tf(A, B, C, D, input=input))


def cont2discrete(sys, dt, method="zoh", alpha=None):  # noqa: C901
    """
    Transform a continuous to a discrete state-space system.

    Parameters
    ----------
    sys : a tuple describing the system.
        The following gives the number of elements in the tuple and
        the interpretation:

           * 2: (num, den)
           * 3: (zeros, poles, gain)
           * 4: (A, B, C, D)

    dt : float
        The discretization time step.
    method : {"gbt", "bilinear", "euler", "backward_diff", "zoh"}
        Which method to use:

           * gbt: generalized bilinear transformation
           * bilinear: Tustin's approximation ("gbt" with alpha=0.5)
           * euler: Euler (or forward differencing) method ("gbt" with alpha=0)
           * backward_diff: Backwards differencing ("gbt" with alpha=1.0)
           * zoh: zero-order hold (default)

    alpha : float within [0, 1]
        The generalized bilinear transformation weighting parameter, which
        should only be specified with method="gbt", and is ignored otherwise

    Returns
    -------
    sysd : tuple containing the discrete system
        Based on the input type, the output will be of the form

        * (num, den, dt)   for transfer function input
        * (zeros, poles, gain, dt)   for zeros-poles-gain input
        * (A, B, C, D, dt) for state-space system input

    Notes
    -----
    By default, the routine uses a Zero-Order Hold (zoh) method to perform
    the transformation.  Alternatively, a generalized bilinear transformation
    may be used, which includes the common Tustin's bilinear approximation,
    an Euler's method technique, or a backwards differencing technique.

    The Zero-Order Hold (zoh) method is based on [1]_, the generalized bilinear
    approximation is based on [2]_ and [3]_.

    References
    ----------
    .. [1] https://en.wikipedia.org/wiki/Discretization
           #Discretization_of_linear_state_space_models

    .. [2] http://techteach.no/publications/discretetime_signals_systems/discrete.pdf

    .. [3] G. Zhang, X. Chen, and T. Chen, Digital redesign via the generalized
           bilinear transformation, Int. J. Control, vol. 82, no. 4, pp. 741-754, 2009.
    """
    if len(sys) == 2:
        sysd = cont2discrete(tf2ss(sys[0], sys[1]), dt, method=method, alpha=alpha)
        return ss2tf(sysd[0], sysd[1], sysd[2], sysd[3]) + (dt,)
    elif len(sys) == 3:
        sysd = cont2discrete(
            zpk2ss(sys[0], sys[1], sys[2]), dt, method=method, alpha=alpha
        )
        return ss2zpk(sysd[0], sysd[1], sysd[2], sysd[3]) + (dt,)
    elif len(sys) == 4:
        a, b, c, d = sys
    else:
        raise ValueError(
            "First argument must either be a tuple of 2 (tf), "
            "3 (zpk), or 4 (ss) arrays."
        )

    if method == "gbt":
        if alpha is None:
            raise ValueError(
                "Alpha parameter must be specified for the "
                "generalized bilinear transform (gbt) method"
            )
        elif alpha < 0 or alpha > 1:
            raise ValueError(
                "Alpha parameter must be within the interval "
                "[0,1] for the gbt method"
            )

    if method == "gbt":
        # This parameter is used repeatedly - compute once here
        ima = np.eye(a.shape[0]) - alpha * dt * a
        ad = np.linalg.solve(ima, np.eye(a.shape[0]) + (1.0 - alpha) * dt * a)
        bd = np.linalg.solve(ima, dt * b)

        # Similarly solve for the output equation matrices
        cd = np.linalg.solve(ima.transpose(), c.transpose())
        cd = cd.transpose()
        dd = d + alpha * np.dot(c, bd)

    elif method in ("bilinear", "tustin"):
        return cont2discrete(sys, dt, method="gbt", alpha=0.5)

    elif method in ("euler", "forward_diff"):
        return cont2discrete(sys, dt, method="gbt", alpha=0.0)

    elif method == "backward_diff":
        return cont2discrete(sys, dt, method="gbt", alpha=1.0)

    elif method == "zoh":
        # Build an exponential matrix
        em_upper = np.hstack((a, b))

        # Need to stack zeros under the a and b matrices
        em_lower = np.hstack(
            (np.zeros((b.shape[1], a.shape[0])), np.zeros((b.shape[1], b.shape[1])))
        )

        em = np.vstack((em_upper, em_lower))
        ms = expm(dt * em)

        # Dispose of the lower rows
        ms = ms[: a.shape[0], :]

        ad = ms[:, 0 : a.shape[1]]
        bd = ms[:, a.shape[1] :]

        cd = c
        dd = d

    else:
        raise ValueError(f"Unknown transformation method '{method}'")

    return ad, bd, cd, dd, dt



from nengo.base import NengoObject, ObjView, ProcessParam
from nengo.dists import DistOrArrayParam, ScatteredHypersphere, Uniform
from nengo.exceptions import ReadonlyError
from nengo.neurons import LIF, Direct, NeuronTypeParam
from nengo.params import BoolParam, Default, IntParam, NumberParam


class Ensemble(NengoObject):
    """A group of neurons that collectively represent a vector.

    Parameters
    ----------
    n_neurons : int
        The number of neurons.
    dimensions : int
        The number of representational dimensions.

    radius : int, optional
        The representational radius of the ensemble.
    encoders : Distribution or (n_neurons, dimensions) array_like, optional
        The encoders used to transform from representational space
        to neuron space. Each row is a neuron's encoder; each column is a
        representational dimension.
    intercepts : Distribution or (n_neurons,) array_like, optional
        The point along each neuron's encoder where its activity is zero. If
        ``e`` is the neuron's encoder, then the activity will be zero when
        ``dot(x, e) <= c``, where ``c`` is the given intercept.
    max_rates : Distribution or (n_neurons,) array_like, optional
        The activity of each neuron when the input signal ``x`` is magnitude 1
        and aligned with that neuron's encoder ``e``;
        i.e., when ``dot(x, e) = 1``.
    eval_points : Distribution or (n_eval_points, dims) array_like, optional
        The evaluation points used for decoder solving, spanning the interval
        (-radius, radius) in each dimension, or a distribution from which
        to choose evaluation points.
    n_eval_points : int, optional
        The number of evaluation points to be drawn from the ``eval_points``
        distribution. If None, then a heuristic is used to determine
        the number of evaluation points.
    neuron_type : `~nengo.neurons.NeuronType`, optional
        The model that simulates all neurons in the ensemble
        (see `~nengo.neurons.NeuronType`).
    gain : Distribution or (n_neurons,) array_like
        The gains associated with each neuron in the ensemble. If None, then
        the gain will be solved for using ``max_rates`` and ``intercepts``.
    bias : Distribution or (n_neurons,) array_like
        The biases associated with each neuron in the ensemble. If None, then
        the gain will be solved for using ``max_rates`` and ``intercepts``.
    noise : Process, optional
        Random noise injected directly into each neuron in the ensemble
        as current. A sample is drawn for each individual neuron on
        every simulation step.
    normalize_encoders : bool, optional
        Indicates whether the encoders should be normalized.
    label : str, optional
        A name for the ensemble. Used for debugging and visualization.
    seed : int, optional
        The seed used for random number generation.

    Attributes
    ----------
    bias : Distribution or (n_neurons,) array_like or None
        The biases associated with each neuron in the ensemble.
    dimensions : int
        The number of representational dimensions.
    encoders : Distribution or (n_neurons, dimensions) array_like
        The encoders, used to transform from representational space
        to neuron space. Each row is a neuron's encoder, each column is a
        representational dimension.
    eval_points : Distribution or (n_eval_points, dims) array_like
        The evaluation points used for decoder solving, spanning the interval
        (-radius, radius) in each dimension, or a distribution from which
        to choose evaluation points.
    gain : Distribution or (n_neurons,) array_like or None
        The gains associated with each neuron in the ensemble.
    intercepts : Distribution or (n_neurons) array_like or None
        The point along each neuron's encoder where its activity is zero. If
        ``e`` is the neuron's encoder, then the activity will be zero when
        ``dot(x, e) <= c``, where ``c`` is the given intercept.
    label : str or None
        A name for the ensemble. Used for debugging and visualization.
    max_rates : Distribution or (n_neurons,) array_like or None
        The activity of each neuron when ``dot(x, e) = 1``,
        where ``e`` is the neuron's encoder.
    n_eval_points : int or None
        The number of evaluation points to be drawn from the ``eval_points``
        distribution. If None, then a heuristic is used to determine
        the number of evaluation points.
    n_neurons : int or None
        The number of neurons.
    neuron_type : NeuronType
        The model that simulates all neurons in the ensemble
        (see ``nengo.neurons``).
    noise : Process or None
        Random noise injected directly into each neuron in the ensemble
        as current. A sample is drawn for each individual neuron on
        every simulation step.
    radius : int
        The representational radius of the ensemble.
    seed : int or None
        The seed used for random number generation.
    """

    probeable = ("decoded_output", "input", "scaled_encoders")

    n_neurons = IntParam("n_neurons", low=1)
    dimensions = IntParam("dimensions", low=1)
    radius = NumberParam("radius", default=1.0, low=1e-10)
    encoders = DistOrArrayParam(
        "encoders",
        default=ScatteredHypersphere(surface=True),
        sample_shape=("n_neurons", "dimensions"),
    )
    intercepts = DistOrArrayParam(
        "intercepts",
        default=Uniform(-1.0, 0.9),
        optional=True,
        sample_shape=("n_neurons",),
    )
    max_rates = DistOrArrayParam(
        "max_rates",
        default=Uniform(200, 400),
        optional=True,
        sample_shape=("n_neurons",),
    )
    eval_points = DistOrArrayParam(
        "eval_points", default=ScatteredHypersphere(), sample_shape=("*", "dimensions")
    )
    n_eval_points = IntParam("n_eval_points", default=None, optional=True)
    neuron_type = NeuronTypeParam("neuron_type", default=LIF())
    gain = DistOrArrayParam(
        "gain", default=None, optional=True, sample_shape=("n_neurons",)
    )
    bias = DistOrArrayParam(
        "bias", default=None, optional=True, sample_shape=("n_neurons",)
    )
    noise = ProcessParam("noise", default=None, optional=True)
    normalize_encoders = BoolParam("normalize_encoders", default=True, optional=True)

    _param_init_order = ["n_neurons", "dimensions"]

    def __init__(
        self,
        n_neurons,
        dimensions,
        radius=Default,
        encoders=Default,
        intercepts=Default,
        max_rates=Default,
        eval_points=Default,
        n_eval_points=Default,
        neuron_type=Default,
        gain=Default,
        bias=Default,
        noise=Default,
        normalize_encoders=Default,
        label=Default,
        seed=Default,
    ):
        super().__init__(label=label, seed=seed)
        self.n_neurons = n_neurons
        self.dimensions = dimensions
        self.radius = radius
        self.encoders = encoders
        self.intercepts = intercepts
        self.max_rates = max_rates
        self.n_eval_points = n_eval_points
        self.eval_points = eval_points
        self.bias = bias
        self.gain = gain
        self.neuron_type = neuron_type
        self.noise = noise
        self.normalize_encoders = normalize_encoders

    def __getitem__(self, key):
        return ObjView(self, key)

    def __len__(self):
        return self.dimensions

    @property
    def neurons(self):
        """A direct interface to the neurons in the ensemble."""
        return Neurons(self)

    @neurons.setter
    def neurons(self, dummy):
        raise ReadonlyError(attr="neurons", obj=self)

    @property
    def size_in(self):
        """The dimensionality of the ensemble."""
        return self.dimensions

    @property
    def size_out(self):
        """The dimensionality of the ensemble."""
        return self.dimensions


class Neurons:
    """An interface for making connections directly to an ensemble's neurons.

    This should only ever be accessed through the ``neurons`` attribute of an
    ensemble, as a way to signal to `~nengo.Connection` that the connection
    should be made directly to the neurons rather than to the ensemble's
    decoded value, e.g.:

    .. testcode::

       with nengo.Network():
           a = nengo.Ensemble(10, 1)
           b = nengo.Ensemble(10, 1)
           nengo.Connection(a.neurons, b.neurons)

    """

    def __init__(self, ensemble):
        self._ensemble = ensemble

    def __getitem__(self, key):
        return ObjView(self, key)

    def __len__(self):
        return self.ensemble.n_neurons

    def __repr__(self):
        return f"<Neurons at 0x{id(self):x} of {self.ensemble!r}>"

    def __str__(self):
        return f"<Neurons of {self.ensemble}>"

    def __eq__(self, other):
        return type(self) == type(other) and self.ensemble is other.ensemble

    def __hash__(self):
        return hash(self.ensemble) + 1  # +1 to avoid collision with ensemble

    @property
    def ensemble(self):
        """(Ensemble) The ensemble these neurons are part of."""
        return self._ensemble

    @property
    def probeable(self):
        """(tuple) Signals that can be probed in the neuron population."""
        return self.ensemble.neuron_type.probeable + ("input",)

    @property
    def size_in(self):
        """(int) The number of neurons in the population."""
        if isinstance(self.ensemble.neuron_type, Direct):
            # This will prevent users from connecting/probing Direct neurons
            # (since there aren't actually any neurons being simulated).
            return 0
        return self.ensemble.n_neurons

    @property
    def size_out(self):
        """(int) The number of neurons in the population."""
        if isinstance(self.ensemble.neuron_type, Direct):
            # This will prevent users from connecting/probing Direct neurons
            # (since there aren't actually any neurons being simulated).
            return 0
        return self.ensemble.n_neurons
